;=-------------------------------------------------------------------------=
;
;			Executable Library Manager v1.0c
;
;     Copyright (c) 1999-2000 by Thomas Kjoernes, all rights reserved.
;
;		Usage:	EXELIB <Library[.EXE]> [object(s)]
;
;=-------------------------------------------------------------------------=

	INCLUDE std.inc
	INCLUDE	dos.inc
	INCLUDE	mzexe.inc
	INCLUDE	mzlib.inc

	.386p

codeSegment SEGMENT USE16

	ASSUME	cs:codeSegment, ds:codeSegment, es:NOTHING

mainEntryPoint:

	; setup our own private safe stack

	push	cs
	pop	ss

	mov	sp, OFFSET ourStackTop

	; clear our uninitialized data area

	push	cs
	pop	es

	mov	di, OFFSET ourClearDef
	mov	cx, (ourClearEnd-ourClearDef+1)/2
	xor	ax, ax
	rep	stosw

	; parse the command line and copy each parameter
	; into a list of null terminated strings

	mov	di, OFFSET ourCmdLine
	call	stdParseCmdLine

	push	cs
	pop	ds

	; set DTA for FindFirst/FindNext operations

	mov	dx, OFFSET ourSearchData
	call	stdSetDTA

	; print welcome banner and quit with an error message
	; if there was absolutely nothing on the command line

	mov	si, OFFSET msgInfo
	call	dosPrintMsgZ

	mov	si, OFFSET msgHelp
	jcxz	errTerminate

	; execute the main piece of code which starts
	; by initializing the specified library

	mov	si, OFFSET ourCmdLine
	call	ourLibraryHandler
	jnc	dosTerminate

errTerminate:
	call	dosPrintMsgZ

dosTerminate:
	mov	ax, 4C00h
	int	21h

;---------------------------------------------------------------------------
;	stdCopyString - Copies a null terminated string into a buffer.
;
;	Entry:
;
;	     ds:si = pointer to source string
;	     es:di = pointer to target buffer
;
;	Exit:
;
;	     es:di = points to null terminator
;
stdCopyString PROC NEAR

	push	ax
	push	si

stdCopyStrNext:
	lodsb			; 
	stosb			; copy string
	or	al, al		;
	jnz	stdCopyStrNext	; until null terminator is found

	dec	di		; point back to NULL

	pop	si
	pop	ax
	ret

stdCopyString ENDP

;---------------------------------------------------------------------------
;	stdSkipString - 
;
;	Entry:
;
stdSkipString PROC NEAR

stdSkipStrMore:	
	lodsb			; skip all characters
	or	al, al		; until null terminator is found
	jnz	stdSkipStrMore	;

	cmp	[si], al	; check if next character is 
	stc			; the string list terminator
	je	stdSkipStrLast	;

	clc			; return success (not last string)

stdSkipStrLast:
	ret

stdSkipString ENDP

;---------------------------------------------------------------------------
;	stdParseCmdLine - 
;
;	Entry:
;		ds = PSP segment
;	     es:di = buffer for "formatted" parameters
;
;	Exit:
;
stdParseCmdLine PROC NEAR

	push	ax
	push	si
	push	di

	xor	cx, cx		; clear argument counter
	mov	si, 128 + 1	; point to command line string

stdParseSkip:
	lodsb			; get command line character

	cmp	al, 13		; terminating CR?
	je	stdParseExit	;
	cmp	al, " "		; space?
	je	stdParseSkip	;

	inc	cx

stdParseMore:
	cmp	al, "a"		; is it too low for conversion?
	jb	stdParseSave	;
	cmp	al, "z"		; is it too high for conversion?
	ja	stdParseSave	;
	sub	al, " "		; in range a..z, make it A..Z

stdParseSave:
	stosb			; copy this character
	lodsb			; also get the next one

	cmp	al, 13		; terminating CR?
	je	stdParseExit	;
	cmp	al, " "		; space?
	jne	stdParseMore	;

	xor	al, al		; we have another string... store
	stosb			; null terminator for previous string
	jmp	stdParseSkip	; and continue with next string

stdParseExit:
	xor	ax, ax		; store double null terminator for
	stosw			; entire list of strings

	pop	di
	pop	si
	pop	ax
	ret

stdParseCmdLine ENDP

;---------------------------------------------------------------------------
;	stdGetPathString - Copies the path to a buffer, excluding filename.
;
;	Entry:
;
;	     ds:si = pointer to source string
;	     es:di = pointer to target buffer
;
;	Exit:
;
;	     es:di = points to null terminator
;
stdGetPathString PROC NEAR

	push	ax
	push	bx
	push	cx
	push	si

	cmp	BYTE PTR [si][1], ":"
	jne	stdGetPathStrMain

	movsw			; store drive specifier
	
stdGetPathStrMain:
	mov	bx, si		; save start pointer
	xor	cx, cx		; zero "\" pointer

stdGetPathStrNext:
	lodsb
	or	al, al
	jz	stdGetPathStrCopy
	cmp	al, "\"
	jne	stdGetPathStrNext

	; we have a directory separator, update "\" pointer

	mov	cx, si
	jmp	stdGetPathStrNext

stdGetPathStrCopy:
	jcxz	stdGetPathStrExit

	mov	si, bx		; pointer start of string
	sub	cx, bx		; calculate path length
	rep	movsb		; copy it!

stdGetPathStrExit:
	xor	al, al		;
	stosb			; store null terminator
	dec	di		; point back to NULL

	pop	si
	pop	cx
	pop	bx
	pop	ax
	ret

stdGetPathString ENDP

;---------------------------------------------------------------------------
;	stdGetNameString - Copies the filename string to a buffer, 
;			   excluding the extension and directory path.
;
;	Entry:
;
;	     ds:si = pointer to source string
;	     es:di = pointer to target buffer
;
;	Exit:
;
;	     es:di = points to null terminator
;
stdGetNameString PROC NEAR

	push	ax
	push	cx
	push	si

	cmp	BYTE PTR [si][1], ":"
	jne	stdGetNameStrMain

	lodsw			; skip drive specifier
	
stdGetNameStrMain:
	mov	bx, si		; save start pointer
	xor	cx, cx		; zero "\" pointer

stdGetNameStrNext:
	lodsb
	or	al, al
	jz	stdGetNameStrCopy
	cmp	al, "."
	je	stdGetNameStrCopy
	cmp	al, "\"
	jne	stdGetNameStrNext

	; we have a directory separator, update "\" pointer

	mov	cx, si
	jmp	stdGetNameStrNext

stdGetNameStrCopy:
	jcxz	stdGetNameStrExit

	xchg	cx, si		; swap "\" and current pointer
	sub	cx, si		; calculate name length
	dec	cx		; minus null/dot
	rep	movsb		; copy it!

stdGetNameStrExit:
	xor	al, al		;
	stosb			; store null terminator
	dec	di		; point back to NULL

	pop	si
	pop	cx
	pop	ax
	ret

stdGetNameString ENDP

;---------------------------------------------------------------------------
;	stdCopyObject - 
;
;	Entry:
;
stdCopyObject PROC NEAR

	pusha

	; attempt to open "object" in read only mode

	mov	al, OPEN_DENY_ALL + OPEN_READ_ONLY
	mov	dx, si
	call	stdFileOpen
	jc	stdCopyObjExit

	; successful -- increment object counter and store handle

	inc	ourLibHeader.mzLibCount
	mov	ourObjHandle, ax
	xchg	ax, bp

	; clear name field first, then get the filename portion,
	; without the full path and extension into there...

	xor	ax, ax
	mov	cx, 4
	mov	di, OFFSET ourObjRecord.mzObjName
	rep	stosw

	mov	di, OFFSET ourObjRecord.mzObjName
	call	stdGetNameString

	; seek to the end of library and save the returned
	; filesize in the object record base field as the position

	mov	bx, ourLibHandle
	call	stdFileSeekEnd

	mov	WORD PTR ourObjRecord.mzObjBase[0], ax
	mov	WORD PTR ourObjRecord.mzObjBase[2], dx

	mov	ourObjRecord.mzObjSize, 0

	; copy the entire object to the end of the library,
	; also increment the object size field on each round

stdCopyObjNext:
	push	bx
	mov	bx, bp
	mov	cx, SIZE ourCopyBuf
	mov	dx, OFFSET ourCopyBuf
	call	stdFileRead
	pop	bx

	add	WORD PTR ourObjRecord.mzObjSize[0], ax
	adc	WORD PTR ourObjRecord.mzObjSize[2], 0

	xchg	ax, cx
	jcxz	stdCopyObjDone

	call	stdFileWrite
	jmp	stdCopyObjNext
	
stdCopyObjDone:	
	
	; realign the library file and close the object

	call	stdFileAlign

	mov	bx, ourObjHandle
	call	stdFileClose

	; write out the new header to the temporary file
	; which contains the library directory

	mov	bx, ourTmpHandle
	mov	cx, SIZE ourObjRecord
	mov	dx, OFFSET ourObjRecord
	call	stdFileWrite

	clc

stdCopyObjExit:
	popa
	ret

stdCopyObject ENDP

;---------------------------------------------------------------------------
;	ourObjectHandler - 
;
;	Entry:
;
ourObjectHandler PROC NEAR

	pusha

	; call the truename function to canonicalize the pathname,
	; this involves expanding any ".." occurancies and incomplete
	; path elements -- very handy.

	mov	di, OFFSET ourTmpPathName
	call	stdTrueName

	; store just the directory path component without the filename,
	; which will change on each FindFirst/FindNext

	mov	si, di
	mov	di, OFFSET ourObjPathName
	call	stdGetPathString

	mov	dx, si
	call	stdFindFirst
	jc	ourObjExit

	; repeat the below steps for each matching file we find,
	; basically say: "Adding: filename.ext" and then append
	; the object to the library!

ourObjFind:
	push	di
	mov	si, OFFSET ourSearchName
	call	stdCopyString
	pop	di

	mov	si, OFFSET msgAdding
	call	dosPrintMsgZ

	mov	si, OFFSET ourObjPathName
	call	dosPrintMsgZ
	call	dosPrintCRLF
	
	call	stdCopyObject

	mov	dx, OFFSET ourSearchData
	call	stdFindNext
	jnc	ourObjFind

ourObjExit:
	popa
	ret

ourObjectHandler ENDP

;---------------------------------------------------------------------------
;	stdInitLibrary - 
;
;	Entry:
;
;	     ds:si = pointer library pathname
;	
;	Exit:
;
stdInitLibrary PROC NEAR

	pushad

	; give us a nice library path

	mov	di, OFFSET ourLibPathName
	call	stdTrueName

	; open library in read/write mode and save handle if
	; we opened it successfully

	mov	al, OPEN_DENY_ALL + OPEN_READ_WRITE
	mov	dx, di
	call	stdFileOpen
	jc	stdInitLibExit

	mov	ourLibHandle, ax
	xchg	ax, bx

	; create a temporary file by the name of "EXELIB.$$$",
	; simply in the current directory... no error checking is
	; done on this one, but that should be added.

	xor	cx, cx
	mov	dx, OFFSET ourTmpFileName
	call	stdFileCreate

	mov	ourTmpHandle, ax
	xchg	ax, bp

	; read first chunk of library, just to verify it
	; as a valid DOS 2.0+ "MZ" executable file

	mov	cx, SIZE ourTempBuf
	mov	dx, OFFSET ourTempBuf
	call	stdFileRead

	cmp	WORD PTR ourTempBuf, "ZM"
	jne	stdInitLibFail

	mov	dx, WORD PTR ourTempBuf.mzExeLibHeader[0]
	mov	cx, WORD PTR ourTempBuf.mzExeLibHeader[2]
	call	stdFileSeek

	push	dx
	push	ax
	pop	edi

	mov	cx, SIZE MZ_LIB
	mov	dx, OFFSET ourTempBuf
	call	stdFileRead

	test	ax, ax
	jz	stdInitLibCreate

	cmp	ourTempBuf.mzLibSignature, MZ_LIB_MAGIC
	jne	stdInitLibCreate

	cmp	edi, ourTempBuf.mzLibStart
	je	stdInitLibUpdate

	; we are creating a new library, simply align the file for now.

stdInitLibCreate:
	call	stdFileAlign

	mov	ax, OFFSET msgCreating
	jmp	stdInitLibDone

	; we are updating the library, copy the existing directory 
	; into the temporary file and truncate the library.

stdInitLibUpdate:
	push	edi	; save header pointer

	; do we have any directory entries to actually copy?

	mov	ecx, ourTempBuf.mzLibCount
	mov	ourLibHeader.mzLibCount, ecx
	jecxz	stdInitLibTruncate

	; OK, we have directory entries, just copy
	; all of them to the temporary file

stdInitLibCopy:
	push	bx
	push	ecx

	mov	cx, SIZE ourObjRecord
	mov	dx, OFFSET ourObjRecord
	call	stdFileRead

	mov	bx, bp
	call	stdFileWrite

	pop	ecx
	pop	bx

	dec	ecx
	jnz	stdInitLibCopy

	; chop the file off at the old library header position

stdInitLibTruncate:
	pop	dx		; restore old header position!
	pop	cx		;

	call	stdFileSeek	; seek there
				
	xor	cx, cx		; writing NULL bytes will make
	call	stdFileWrite	; DOS truncate the file for us

	mov	ax, OFFSET msgUpdating

stdInitLibDone:
	xchg	ax, si		; print message (pointed to by ax)
	call	dosPrintMsgZ	;
	xchg	ax, si		; print library filename
	call	dosPrintMsgZ	;
	call	dosPrintCRLF	; print CR/LF for neatness

	clc
	jmp	stdInitLibExit

stdInitLibFail:
	call	stdFileClose	; we failed to init library, close
	mov	bx, bp		; both the library and our temporary
	call	stdFileClose	; directory file
	stc

stdInitLibExit:
	popad
	ret

stdInitLibrary ENDP

;---------------------------------------------------------------------------
;	stdDoneLibrary - 
;
;	Entry:
;	
;	Exit:
;
stdDoneLibrary PROC NEAR

	pusha

	mov	bx, ourLibHandle	; align library
	call	stdFileAlign		;

	call	stdFileSeekEnd

	push	dx			; save estimated library
	push	ax			; header position...

	; read EXE header and save the library "hint" pointer
	; at offset 38h in the DOS header...
	
	call	stdFileSeekStart	; seek to start of file
	mov	cx, SIZE MZ_EXE		;
	mov	dx, OFFSET ourCopyBuf	; read "MZ" header
	call	stdFileRead		;

	pop	eax

	; is it safe to store hint pointer?

	cmp	ourCopyBuf.mzRelocTable, 3Ch
	jb	stdDoneLibSkip

	mov	ourCopyBuf.mzExeLibHeader, eax
	
	call	stdFileSeekStart	; seek to start again
	mov	cx, SIZE MZ_EXE		;
	mov	dx, OFFSET ourCopyBuf	; write "MZ" header
	call	stdFileWrite		;

stdDoneLibSkip:	
	call	stdFileSeekEnd

	; save "estimated" library header position,
	; then write it out to the file

	mov	ourLibHeader.mzLibSignature, MZ_LIB_MAGIC

	mov	WORD PTR ourLibHeader.mzLibStart[0], ax
	mov	WORD PTR ourLibHeader.mzLibStart[2], dx

	mov	cx, SIZE ourLibHeader
	mov	dx, OFFSET ourLibHeader
	call	stdFileWrite

	; seek to start of temporary file -- containing 
	; all of our directory entries!

	mov	bx, ourTmpHandle
	call	stdFileSeekStart

	mov	ecx, ourLibHeader.mzLibCount
	jecxz	stdDoneLibClose

	; copy the entries one by one, this might be
	; just a bit slower than copying big chunks...

stdDoneLibCopy:
	push	ecx

	mov	cx, SIZE ourObjRecord
	mov	dx, OFFSET ourObjRecord
	call	stdFileRead

	push	bx
	mov	bx, ourLibHandle
	call	stdFileWrite
	pop	bx

	pop	ecx
	dec	ecx
	jnz	stdDoneLibCopy

	; close the library and erase the temporary file

stdDoneLibClose:
	call	stdFileClose
	mov	bx, ourLibHandle
	call	stdFileClose

	mov	dx, OFFSET ourTmpFileName
	call	stdFileErase

	popa
	ret

stdDoneLibrary ENDP

;---------------------------------------------------------------------------
;	ourLibraryHandler - 
;
;	Entry:
;
ourLibraryHandler PROC NEAR

	; attempt to initialize the library whose 
	; complete path is pointed to by ds:si

	call	stdInitLibrary
	jnc	ourLibMain

	; exit with an error message being shown

	push	si
	mov	si, OFFSET msgErrLibInit
	call	dosPrintMsgZ
	pop	si
	call	dosPrintMsgZ
	call	dosPrintCRLF
	jmp	ourLibExit

	; skip the library argument and call the object handler
	; for each string argument that were on the command line

ourLibMain:
	call	stdSkipString
	jc	ourLibDone

ourLibNext:
	call	ourObjectHandler
	call	stdSkipString
	jnc	ourLibNext

ourLibDone:
	call	stdDoneLibrary	; clean up after us
	clc

ourLibExit:
	ret

ourLibraryHandler ENDP

;---------------------------------------------------------------------------
;	stdSetDTA - 
;
;	Entry:
;	
;	Exit:
;
stdSetDTA PROC NEAR

	mov	ah, 1Ah
	int	21h
	ret

stdSetDTA ENDP

;---------------------------------------------------------------------------
;	stdFileCreate - 
;
;	Entry:
;	
;	Exit:
;
stdFileCreate PROC NEAR

	mov	ah, 3Ch
	int	21h
	ret

stdFileCreate ENDP

;---------------------------------------------------------------------------
;	stdFileOpen - 
;
;	Entry:
;	
;	Exit:
;
stdFileOpen PROC NEAR

	mov	ah, 3Dh
	int	21h
	ret

stdFileOpen ENDP

;---------------------------------------------------------------------------
;	stdFileClose - 
;
;	Entry:
;	
;	Exit:
;
stdFileClose PROC NEAR

	mov	ah, 3Eh
	int	21h
	ret

stdFileClose ENDP

;---------------------------------------------------------------------------
;	stdFileRead - 
;
;	Entry:
;	
;	Exit:
;
stdFileRead PROC NEAR

	mov	ah, 3Fh
	int	21h
	ret

stdFileRead ENDP

;---------------------------------------------------------------------------
;	stdFileWrite - 
;
;	Entry:
;	
;	Exit:
;
stdFileWrite PROC NEAR

	mov	ah, 40h
	int	21h
	ret

stdFileWrite ENDP

;---------------------------------------------------------------------------
;	stdFileErase - 
;
;	Entry:
;	
;	Exit:
;
stdFileErase PROC NEAR
	
	mov	ah, 41h
	int	21h
	ret

stdFileErase ENDP

;---------------------------------------------------------------------------
;	stdFileSeekEnd - 
;
;	Entry:
;	
;	Exit:
;
stdFileSeekEnd PROC NEAR

	mov	ax, 4202h
	xor	cx, cx
	xor	dx, dx
	int	21h
	ret

stdFileSeekEnd ENDP

;---------------------------------------------------------------------------
;	stdFileSeekStart - 
;
;	Entry:
;	
;	Exit:
;
stdFileSeekStart PROC NEAR

	xor	cx, cx
	xor	dx, dx

stdFileSeekStart ENDP

;---------------------------------------------------------------------------
;	stdFileSeek - 
;
;	Entry:
;	
;	Exit:
;
stdFileSeek PROC NEAR

	mov	ax, 4200h
	int	21h
	ret

stdFileSeek ENDP

;---------------------------------------------------------------------------
;	stdFileAlign - 
;
;	Entry:
;	
;	Exit:
;
stdFileAlign PROC NEAR

	pusha

	call	stdFileSeekEnd		; get file size

	and	ax, MZ_OBJ_ALIGN - 1	; keep alignment bits
	jz	stdFileAlreadyAligned	; zero?

	mov	cx, MZ_OBJ_ALIGN	; find out how many bytes
	sub	cx, ax			; missing before aligned

	mov	dx, OFFSET ourZeroBuf	; write them out to the
	call	stdFileWrite		; file as zero-bytes

stdFileAlreadyAligned:
	popa
	ret

stdFileAlign ENDP

;---------------------------------------------------------------------------
;	stdFindFirst - 
;
;	Entry:
;	
;	Exit:
;
stdFindFirst PROC NEAR

	mov	ah, 4Eh
	int	21h
	ret

stdFindFirst ENDP

;---------------------------------------------------------------------------
;	stdFindNext - 
;
;	Entry:
;	
;	Exit:
;
stdFindNext PROC NEAR

	mov	ah, 4Fh
	int	21h
	ret

stdFindNext ENDP

;---------------------------------------------------------------------------
;	stdTrueName - 
;
;	Entry:
;	
;	Exit:
;
stdTrueName PROC NEAR
	
	mov	ah, 60h
	int	21h
	ret

stdTrueName ENDP

;---------------------------------------------------------------------------
;	dosPrintChar -
;
;	Entry:
;		al = character
;
dosPrintChar PROC NEAR

	push	dx
	mov	ah, 02h
	mov	dl, al
	int	21h
	pop	dx
	ret

dosPrintChar ENDP

;---------------------------------------------------------------------------
;	dosPrintCRLF -
;
;	Entry:
;		None
;
dosPrintCRLF PROC NEAR

	push	ax
	mov	al, 13		; output ASCII 13 - CR
	call	dosPrintChar	;
	mov	al, 10		; output ASCII 10 - LF
	call	dosPrintChar	;
	pop	ax
	ret

dosPrintCRLF ENDP

;---------------------------------------------------------------------------
;	dosPrintMsgZ -
;
;	Entry:
;
;	     ds:si = pointer to message
;
dosPrintMsgZ PROC NEAR

	push	ax
	push	si

dosPMZNext:
	lodsb			;
	or	al, al		;
	jz	dosPMZDone	;
	call	dosPrintChar	;
	jmp	dosPMZNext	;

dosPMZDone:
	pop	si
	pop	ax
	ret

dosPrintMsgZ ENDP

;---------------------------------------------------------------------------
; Status Messages
;---------------------------------------------------------------------------

msgInfo		DB	"Executable Library Manager v1.0c [RM] ", BUILD, 13, 10
		DB	"Copyright (c) 1999-2000 by Thomas Kjoernes, "
		DB	"all rights reserved.", 13, 10
		DB	13, 10, "  ", NULL

msgHelp		DB	"Usage: EXELIB <Library[.EXE]> [object(s)]"
		DB	13, 10, NULL

msgCRLF		DB	13, 10, NULL

msgCreating	DB	13, "  Creating LIB: ", NULL
msgUpdating	DB	13, "  Updating LIB: ", NULL

msgAdding	DB	"    Adding ", NULL

msgErrLibInit	DB	"  Error initializing library - ", NULL

;---------------------------------------------------------------------------
; Data Area
;---------------------------------------------------------------------------

ourTmpFileName	DB	"EXELIB.$$$", NULL

ourClearDef	LABEL	WORD

ourLibHeader	MZ_LIB	<>
ourObjRecord	MZ_OBJ	<>

ourLibPathName	DB	80 DUP (?)
ourObjPathName	DB	80 DUP (?)
ourTmpPathName	DB	80 DUP (?)

ourSearchData	DB	21 DUP (?)
ourSearchAttr	DB	?
ourSearchTime	DW	?
ourSearchDate	DW	?
ourSearchSize	DD	?
ourSearchName	DB	12 DUP (?)

ourCmdLine	DB	128 DUP (?)

ourTmpHandle	DW	?
ourLibHandle	DW	?
ourObjHandle	DW	?

ourTempBuf	DB	MZ_OBJ_ALIGN DUP (?)
ourZeroBuf	DB	MZ_OBJ_ALIGN DUP (?)

ourCopyBuf	DB	32768 DUP (?)

ourStackDef 	LABEL 	WORD
		DW	1024 DUP (?)
ourStackTop 	LABEL 	WORD

ourClearEnd	LABEL	WORD

;---------------------------------------------------------------------------

codeSegment ENDS

	END	mainEntryPoint
