;---------------------------------------------------------------------------
;
; 		       8042 Keyboard Controller Routines
;
;     Copyright (c) 1999-2001 by Thomas Kjoernes, all rights reserved.
;
;	
;	API Routines:
;
;	- ISA_KBC_SendCommand
;	- ISA_KBC_SendData
;	- ISA_KBC_ReceiveData
;	- ISA_KBC_SendToKbd
;	- ISA_KBC_ReceiveFromKbd
;	- ISA_KBC_SendToAux
;	- ISA_KBC_ReceiveFromAux
;
;---------------------------------------------------------------------------

	INCLUDE std.inc
	INCLUDE stdgroup.inc

	.386p

	PUBLIC	kbcSendCommand
	PUBLIC	kbcSendData
	PUBLIC	kbcReceiveData
	PUBLIC	kbcSendToKbd
	PUBLIC	kbcReceiveFromKbd
	PUBLIC	kbcSendToAux
	PUBLIC	kbcReceiveFromAux

codeSegment SEGMENT

;---------------------------------------------------------------------------
;	kbcSendCommand
;
;	Entry:
;		al = Controller Command
;	Exit:
;		CF==0 if Success
;		CF==1 if Error
;
DefProc kbcSendCommand

	push	eax
	push	ecx

	mov	ah, al

	mov	ecx, 00080000h

kbcSendCommandWait:
	in	al, 64h
	test	al, 00000010b
	loopnz	kbcSendCommandWait

	stc
	jnz	kbcSendCommandExit

	mov	al, ah
	out	64h, al

	clc

kbcSendCommandExit:
	pop	ecx
	pop	eax
	ret

EndProc kbcSendCommand

;---------------------------------------------------------------------------
;	kbcSendData
;
;	Entry:
;		al = Data
;	Exit:
;		CF==0 if Success
;		CF==1 if Error
;
DefProc kbcSendData

	push	eax
	push	ecx

	mov	ah, al

	mov	ecx, 00080000h

kbcSendDataWait:
	in	al, 64h
	test	al, 00000010b
	loopnz	kbcSendDataWait

	stc
	jnz	kbcSendDataExit

	mov	al, ah
	out	60h, al

	clc

kbcSendDataExit:
	pop	ecx
	pop	eax
	ret

EndProc kbcSendData

;---------------------------------------------------------------------------
;	kbcReceiveData
;
;	Entry:
;		None
;	Exit:
;		CF==0 if Success
;
;		al = Data
;
;		CF==1 if Error
;
DefProc kbcReceiveData

	push	ecx

	mov	ecx, 00100000h

kbcReceiveDataWait:
	in	al, 64h
	test	al, 00000001b
	loopz	kbcReceiveDataWait

	stc
	jz	kbcReceiveDataExit

	in	al, 60h

	clc

kbcReceiveDataExit:
	pop	ecx
	ret

EndProc kbcReceiveData

;---------------------------------------------------------------------------
;	kbcSendToKbd
;
;	Entry:
;		al = Keyboard Command/Data
;	Exit:
;		CF==0 if Success
;		CF==1 if Error
;
DefProc kbcSendToKbd

	call	kbcSendData
	ret

EndProc kbcSendToKbd

;---------------------------------------------------------------------------
;	kbcReceiveFromKbd
;
;	Entry:
;		None
;	Exit:
;		CF==0 if Success
;
;		al = Data
;
;		CF==1 if Error
;
DefProc kbcReceiveFromKbd

	call	kbcReceiveData
	ret

EndProc kbcReceiveFromKbd

;---------------------------------------------------------------------------
;	kbcSendToAux
;
;	Entry:
;		al = Auxiliary Command/Data
;	Exit:
;		CF==0 if Success
;		CF==1 if Error
;
DefProc kbcSendToAux

	push	eax
	mov	al, 0D4h
	call	kbcSendCommand
	pop	eax
	call	kbcSendData
	ret

EndProc kbcSendToAux

;---------------------------------------------------------------------------
;	kbcReceiveFromAux
;
;	Entry:
;		None
;	Exit:
;		CF==0 if Success
;
;		al = Data
;
;		CF==1 if Error
;
DefProc kbcReceiveFromAux

	call	kbcReceiveData
	ret

EndProc kbcReceiveFromAux

codeSegment ENDS

;---------------------------------------------------------------------------
; Data
;---------------------------------------------------------------------------

dataSegment SEGMENT

dataSegment ENDS

;---------------------------------------------------------------------------
; BSS Data
;---------------------------------------------------------------------------

bssSegment SEGMENT

bssSegment ENDS

	END

