;---------------------------------------------------------------------------
;
; 		     	32-bit PM System Kernel Module
;
;     Copyright (c) 1999-2001 by Thomas Kjoernes, all rights reserved.
;
;---------------------------------------------------------------------------

	INCLUDE	std.inc
	INCLUDE	stdgroup.inc
	INCLUDE	kernel.api

	.386p

	; Exports

	PUBLIC	keFatalExit

codeSegment SEGMENT

	; Initialization

	EXTERN	expInit:NEAR
	EXTERN	irqInit:NEAR
	EXTERN	apiInit:NEAR
	EXTERN	kdbInit:NEAR
	EXTERN	kmmInit:NEAR
	EXTERN	tssInit:NEAR
	EXTERN	v86Init:NEAR
	EXTERN	ifsInit:NEAR

	; Shutdown

	EXTERN	expDone:NEAR
	EXTERN	irqDone:NEAR
	EXTERN	apiDone:NEAR
	EXTERN	kdbDone:NEAR
	EXTERN	kmmDone:NEAR
	EXTERN	tssDone:NEAR
	EXTERN	v86Done:NEAR
	EXTERN	ifsDone:NEAR

;---------------------------------------------------------------------------
;	mainEntryPoint - KERNEL32.SYS entry point
;
;
;	Entry (Registers):
;
;		ebx = our memory block base -- "module handle"
;		ecx = our memory block size
;		ebp = pointer to arguments on the stack
;
;	Entry (Stack):
;
;		DWORD = return address
;
;	Entry (Stack Arguments):
;
;		DWORD	Base of Library
;		DWORD	Size of Library
;		DWORD	Base of Reclaim Area
;		DWORD	Size of Reclaim Area
;		DWORD	Base of MDT
;		DWORD	Size of MDT (number of entries)
;		DWORD	Base of PSP
;		DWORD	Base of Environment
;
;	Exit:
;		None
;
DefProc mainEntryPoint

	push	EFLAGS_RESET
	popfd

	mov	ourImageBase, ebx	; save image base
	mov	ourImageSize, ecx	; save image size
	mov	ourImageArgs, ebp	; save image argument ptr

	sidt	loaderIDTPtr		; save old IDT
	sgdt	loaderGDTPtr		; save old GDT

	mov	esi, loaderGDTPtr.gdtrBase
	mov	edi, kernelGDTPtr.gdtrBase
	mov	ecx, 64
	rep	movsb

	lidt	kernelIDTPtr		; load new IDT
	lgdt	kernelGDTPtr		; load new GDT

	pop	loaderReturnOfs		; save return address

	call	expInit
	call	irqInit
	call	apiInit
	call	kdbInit
	call	kmmInit
	call	tssInit
	call	v86Init
	call	ifsInit

	;-------------------------------------------------------------
	; load sub-system modules
	;-------------------------------------------------------------

kernelMainExit:

	cli
	cld

	call	ifsDone
	call	v86Done
	call	tssDone
	call	kmmDone
	call	kdbDone
	call	apiDone
	call	irqDone
	call	expDone

	lgdt	loaderGDTPtr		; restore previous GDT
	lidt	loaderIDTPtr		; restore previous IDT
	
	jmp	loaderReturn		; back...

EndProc mainEntryPoint

;---------------------------------------------------------------------------
;	keGetVersion
;
;	Entry:
;		None
;	Exit:
;		eax = version ID
;
DefProc keGetVersion

	mov	eax, 00000100h
	clc
	ret

EndProc keGetVersion

;---------------------------------------------------------------------------
;	keGetKernelInfo
;
;	Entry:
;		None
;	Exit:
;		esi = address of KERNEL_ARG
;
DefProc keGetKernelInfo

	mov	ebx, ourImageBase
	mov	ecx, ourImageSize
	mov	esi, ourImageArgs
	ret

EndProc keGetKernelInfo

;---------------------------------------------------------------------------
;	keFatalExit - BOOM!
;
;	Entry:
;		None
;	Exit:
;		None
;
DefProc keFatalExit

	; paranoid: reload segment registers

	mov	eax, DATA32

	mov	ss, eax
	mov	ds, eax
	mov	es, eax
	mov	fs, eax

	mov	eax, DATA16
	mov	gs, eax

	jmp	kernelMainExit

EndProc keFatalExit

	; Services

	EXTERN	peLoadModule:NEAR
	EXTERN	peFreeModule:NEAR

	EXTERN	kmmHeapAllocate:NEAR
	EXTERN	kmmHeapFree:NEAR

	EXTERN	gdtGetEntry:NEAR
	EXTERN	gdtSetEntry:NEAR

	EXTERN	idtGetEntry:NEAR
	EXTERN	idtSetEntry:NEAR

	EXTERN	tssGetRing0ESP:NEAR
	EXTERN	tssSetRing0ESP:NEAR

	EXTERN	expGetHandlerPM:NEAR
	EXTERN	expSetHandlerPM:NEAR
	EXTERN	expGetHandlerVM:NEAR
	EXTERN	expSetHandlerVM:NEAR

	EXTERN	irqGetHandlerPM:NEAR
	EXTERN	irqSetHandlerPM:NEAR
	EXTERN	irqGetHandlerVM:NEAR
	EXTERN	irqSetHandlerVM:NEAR
	EXTERN	irqGetPMIntByIRQ:NEAR
	EXTERN	irqGetVMIntByIRQ:NEAR

	EXTERN	picGetBase:NEAR
	EXTERN	picSetBase:NEAR
	EXTERN	picMask:NEAR
	EXTERN	picUnMask:NEAR
	EXTERN	picEOI:NEAR

	EXTERN	apiRegister:NEAR
	EXTERN	apiUnRegister:NEAR
	EXTERN	apiGetProcEntry:NEAR
	EXTERN	apiGetProcAddress:NEAR

	EXTERN	v86GetExitProc:NEAR
	EXTERN	v86SetExitProc:NEAR

	EXTERN	ifsFileOpen:NEAR
	EXTERN	ifsFileClose:NEAR
	EXTERN	ifsFileRead:NEAR
	EXTERN	ifsFileWrite:NEAR
	EXTERN	ifsFileGetPos:NEAR
	EXTERN	ifsFileSetPos:NEAR

	EXTERN	rtlListGetHead:NEAR
	EXTERN	rtlListGetTail:NEAR
	EXTERN	rtlListGetPrev:NEAR
	EXTERN	rtlListGetNext:NEAR
	EXTERN	rtlListAttachHead:NEAR
	EXTERN	rtlListAttachTail:NEAR
	EXTERN	rtlListDetachHead:NEAR
	EXTERN	rtlListDetachTail:NEAR
	EXTERN	rtlListInsert:NEAR
	EXTERN	rtlListDelete:NEAR

codeSegment ENDS

;---------------------------------------------------------------------------
; Data
;---------------------------------------------------------------------------

dataSegment SEGMENT

;---------------------------------------------------------------------------
; API Service Table
;---------------------------------------------------------------------------

	PUBLIC	apiServiceTbl

apiServiceTbl	LABEL	DWORD
	API_TABLE	<apiServiceDef, (apiServiceEnd-apiServiceDef)/8>
apiServiceDef	LABEL	DWORD
	API_ENTRY	<KERNEL_GetVersion,		keGetVersion>
	API_ENTRY	<KERNEL_GetKernelInfo,		keGetKernelInfo>
	API_ENTRY	<KERNEL_FatalExit,		keFatalExit>
	API_ENTRY	<KERNEL_Reboot,			keFatalExit>
	API_ENTRY	<KERNEL_LoadModule,		peLoadModule>
	API_ENTRY	<KERNEL_FreeModule,		peFreeModule>
	API_ENTRY	<KERNEL_HeapAllocate,		kmmHeapAllocate>
	API_ENTRY	<KERNEL_HeapFree,		kmmHeapFree>
	API_ENTRY	<KERNEL_ApiRegister,		apiRegister>
	API_ENTRY	<KERNEL_ApiUnRegister,		apiUnRegister>
	API_ENTRY	<KERNEL_ApiGetProcEntry,	apiGetProcEntry>
	API_ENTRY	<KERNEL_ApiGetProcAddress,	apiGetProcAddress>
	API_ENTRY	<KERNEL_GdtGetEntry,		gdtGetEntry>
	API_ENTRY	<KERNEL_GdtSetEntry,		gdtSetEntry>
	API_ENTRY	<KERNEL_IdtGetEntry,		idtGetEntry>
	API_ENTRY	<KERNEL_IdtSetEntry,		idtSetEntry>
	API_ENTRY	<KERNEL_TssGetRing0ESP,	tssGetRing0ESP>
	API_ENTRY	<KERNEL_TssSetRing0ESP,	tssSetRing0ESP>
	API_ENTRY	<KERNEL_ExpGetHandlerPM,	expGetHandlerPM>
	API_ENTRY	<KERNEL_ExpSetHandlerPM,	expSetHandlerPM>
	API_ENTRY	<KERNEL_ExpGetHandlerVM,	expGetHandlerVM>
	API_ENTRY	<KERNEL_ExpSetHandlerVM,	expSetHandlerVM>
	API_ENTRY	<KERNEL_IrqGetHandlerPM,	irqGetHandlerPM>
	API_ENTRY	<KERNEL_IrqSetHandlerPM,	irqSetHandlerPM>
	API_ENTRY	<KERNEL_IrqGetHandlerVM,	irqGetHandlerVM>
	API_ENTRY	<KERNEL_IrqSetHandlerVM,	irqSetHandlerVM>
	API_ENTRY	<KERNEL_IrqGetPMIntByIRQ,	irqGetPMIntByIRQ>
	API_ENTRY	<KERNEL_IrqGetVMIntByIRQ,	irqGetVMIntByIRQ>
	API_ENTRY	<KERNEL_IrqGetBase,		picGetBase>
	API_ENTRY	<KERNEL_IrqSetBase,		picSetBase>
	API_ENTRY	<KERNEL_IrqMask,		picMask>
	API_ENTRY	<KERNEL_IrqUnMask,		picUnMask>
	API_ENTRY	<KERNEL_IrqEOI,			picEOI>
	API_ENTRY	<KERNEL_V86GetExitProc,		v86GetExitProc>
	API_ENTRY	<KERNEL_V86SetExitProc,		v86SetExitProc>
	API_ENTRY	<KERNEL_IfsFileOpen,		ifsFileOpen>
	API_ENTRY	<KERNEL_IfsFileClose,		ifsFileClose>
	API_ENTRY	<KERNEL_IfsFileRead,		ifsFileRead>
	API_ENTRY	<KERNEL_IfsFileWrite,		ifsFileWrite>
	API_ENTRY	<KERNEL_IfsFileGetPos,		ifsFileGetPos>
	API_ENTRY	<KERNEL_IfsFileSetPos,		ifsFileSetPos>
	API_ENTRY	<KERNEL_ListGetHead,		rtlListGetHead>
	API_ENTRY	<KERNEL_ListGetTail,		rtlListGetTail>
	API_ENTRY	<KERNEL_ListGetPrev,		rtlListGetPrev>
	API_ENTRY	<KERNEL_ListGetNext,		rtlListGetNext>
	API_ENTRY	<KERNEL_ListAttachHead,		rtlListAttachHead>
	API_ENTRY	<KERNEL_ListAttachTail,		rtlListAttachTail>
	API_ENTRY	<KERNEL_ListDetachHead,		rtlListDetachHead>
	API_ENTRY	<KERNEL_ListDetachTail,		rtlListDetachTail>
	API_ENTRY	<KERNEL_ListInsert,		rtlListInsert>
	API_ENTRY	<KERNEL_ListDelete,		rtlListDelete>
apiServiceEnd	LABEL	DWORD

; Our ticket back...

loaderReturn	LABEL	FWORD
loaderReturnOfs	DD	?
loaderReturnSeg	DD	CODE32

; GDT and IDT registers

	EXTERN	kernelGDTPtr:FWORD
	EXTERN	kernelIDTPtr:FWORD

; List of 'static' Kernel Modules

szStaticDef	LABEL	BYTE
	DB	"",	NULL
szStaticEnd	LABEL	BYTE
	DB		NULL

dataSegment ENDS

;---------------------------------------------------------------------------
; BSS Data
;---------------------------------------------------------------------------

bssSegment SEGMENT

; GDT and IDT storage

	EXTERN	kernelGDTDef:DWORD
	EXTERN	kernelGDTEnd:DWORD

	EXTERN	kernelIDTDef:DWORD
	EXTERN	kernelIDTEnd:DWORD

loaderGDTPtr	DQ	?
loaderIDTPtr	DQ	?

ourImageBase	DD	?
ourImageSize	DD	?
ourImageArgs	DD	?

bssSegment ENDS

	END	mainEntryPoint

