;---------------------------------------------------------------------------
;
;		       Operating System Loader v1.1 [RM]
;
;     Copyright (c) 1999-2001 by Thomas Kjoernes, all rights reserved.
;
;---------------------------------------------------------------------------

	.386p

	INCLUDE	std.inc
	INCLUDE	dos.inc
	INCLUDE	xms.inc
	INCLUDE	mzexe.inc
	INCLUDE	mzlib.inc
	INCLUDE	peexe.inc

;---------------------------------------------------------------------------
; 16-bit Real Mode
;---------------------------------------------------------------------------

codeSegment SEGMENT USE16 'code'

	ASSUME	cs:codeSegment, ds:dataSegment

;---------------------------------------------------------------------------
;	bootEntryPoint - MZ Loader 2.0 compatible
;
;	Entry (Registers):
;
;		ds = load segment
;		es = free segment
;		ss = last segment
;
;		dl = boot drive number
;
;    	     ss:sp = pointer to 2KB stack (below top of memory).
;    	     ss:bp = pointer to boot sector, with MZ_LOADER_VARS
;		       located below.
;
;	Exit:
;	     	None
;
bootEntryPoint PROC NEAR

	mov	eax, ds			; eax = "load" address
	shl	eax, 4			;

	mov	ebx, es			; ebx = "free" address
	shl	ebx, 4			;

	push	dataSegment
	pop	ds

	; calculate image size, then save it along with the base

	mov	ecx, ebx
	sub	ecx, eax

	mov	ourLibraryBase, eax
	mov	ourLibrarySize, ecx

	; calculate reclaim size, then save it along with the base

	mov	eax, stackSegment
	shl	eax, 4

	mov	ecx, ebx
	sub	ecx, eax	

	mov	ourReclaimBase, eax
	mov	ourReclaimSize, ecx

	; calculate size of remaining free memory and save it

	mov	ecx, ss
	shl	ecx, 4
	sub	ecx, ebx

	call	rawConvMemory

	mov	si, OFFSET msgLoader
	call	stdPrintMsgZ

	call	ps2EnableA20
	call	kbcEnableA20

	push	ds
	push	es
	push	fs
	push	gs

	call	rawFindMemory
	call	stdLoaderMain

	pop	gs
	pop	fs
	pop	es
	pop	ds

	call	kbcDisableA20
	call	ps2DisableA20

	; notify the user about a very soft reboot...

	mov	si, OFFSET msgReboot
	call	stdPrintMsgZ

	mov	ah, 00h
	int	16h

	call	stdPrintCRLF

	int	19h

bootEntryPoint ENDP

;---------------------------------------------------------------------------
;	mainEntryPoint -- LOADER.EXE
;
;	Entry:
;		ds = PSP segment
;		es = PSP segment
;	Exit:
;	     	None
;
mainEntryPoint PROC NEAR

	; did we come from DOS or boot sector?

	cmp	WORD PTR ds:[0], DOS_PSP_MAGIC
	jne	bootEntryPoint

;	int	3

	mov	dx, ds:[2Ch]

	mov	ax, dataSegment
	mov	ds, ax

	mov	ourPSPSeg, es
	mov	ourEnvSeg, dx

	mov	si, OFFSET msgLoader
	call	stdPrintMsgZ

	mov	si, OFFSET msgErrV86

	smsw	ax
	shr	ax, 1
	jc	mainExitFailure

	; shrink our memory block to the minimum needed

	mov	ah, 4Ah
	mov	bx, lastSegment
	sub	bx, codeSegment
	add	bx, 16
	int	21h

	mov	si, OFFSET msgErrLib

	call	dosLoadLibrary
	jc	mainExitFailure

	mov	ourLibraryBase, eax
	mov	ourLibrarySize, ecx

	mov	ourReclaimBase, eax
	mov	ourReclaimSize, ecx

	call	dosGrabMemory

	call	xmsInitDriver
	jc	mainPureMode

	call	xmsGrabMemory
	call	stdLoaderMain
	call	xmsFreeMemory

	jmp	mainExitSuccess

mainPureMode:
	call	ps2EnableA20
	call	kbcEnableA20

	call	rawFindMemory
	call	stdLoaderMain

	call	kbcDisableA20
	call	ps2DisableA20

mainExitSuccess:
	mov	ax, 4C00h
	int	21h	

mainExitFailure:
	call	stdPrintMsgZ

	mov	ax, 4CFFh
	int	21h	

mainEntryPoint ENDP

;---------------------------------------------------------------------------
;	dosLoadLibrary
;
;	Entry:
;		ds = dataSegment
;	Exit:
;	       eax = executable base address
;	       ecx = executable size
;
dosLoadLibrary PROC NEAR

	push	bx
	push	dx
	push	si
	push	ds
	push	es

	mov	ds, ourEnvSeg
	xor	ax, ax
	xor	si, si

	; find end of environment and our path

dosLoadLibraryFind:
	cmp	ax, ds:[si]
	je	dosLoadLibraryOpen
	inc	si
	jns	dosLoadLibraryFind

dosLoadLibraryFail:
	xor	eax, eax
	xor	ecx, ecx
	stc
	jmp	dosLoadLibraryExit

	; attempt to open file

dosLoadLibraryOpen:
	add	si, 4
	mov	ax, 3D00h
	mov	dx, si
	int	21h
	jc	dosLoadLibraryFail

	xchg	ax, bx

	; determine file size and memory requirements

	mov	ax, 4202h
	xor	cx, cx
	xor	dx, dx
	int	21h

	push	dx
	push	ax

	mov	ax, 4200h
	xor	cx, cx
	xor	dx, dx
	int	21h

	pop	ecx

	; round up and convert to paragraphs

	mov	eax, ecx
	add	eax, 15
	shr	eax, 4

	; attempt to allocate memory

	push	bx
	xchg	ax, bx
	mov	ah, 48h
	int	21h
	pop	bx
	jc	dosLoadLibraryFail

	; load file into memory chunk by chunk

	movzx	eax, ax

	push	eax
	push	ecx

dosLoadLibraryMain:
	push	ax
	mov	ds, ax

	mov	ah, 3Fh
	mov	cx, 32768
	xor	dx, dx
	int	21h

	test	ax, ax
	pop	ax
	jz	dosLoadLibraryDone

	add	ax, 32768 / 16
	jmp	dosLoadLibraryMain

dosLoadLibraryDone:
	mov	ah, 3Eh
	int	21h

	pop	ecx
	pop	eax

	shl	eax, 4
	clc

dosLoadLibraryExit:
	pop	es
	pop	ds
	pop	si
	pop	dx
	pop	bx
	ret

dosLoadLibrary ENDP

;---------------------------------------------------------------------------
;	stdLoaderMain
;
;	Entry:
;		ds = dataSegment
;	Exit:
;		None
;
stdLoaderMain PROC NEAR

	call	stdDebugInit

	cli

	call	stdLoaderPM16

	sti

	call	stdDebugDone

	ret

stdLoaderMain ENDP

;---------------------------------------------------------------------------
;	stdLoaderPM16
;
;	Entry:
;		ds = dataSegment
;	Exit:
;		None
;
stdLoaderPM16 PROC FAR

	mov	oldStackOfs, esp
	mov	oldStackSeg, ss

	; align stack on DWORD boundary and "flatten" ESP

	and	esp, 0FFFCh
	mov	eax, ss
	shl	eax, 4
	add	esp, eax

	; patch 16-bit code selector

	mov	eax, cs
	shl	eax, 4
	or	DWORD PTR loaderGDTPtr[CODE16][2], eax

	; patch 16-bit data selector

	mov	eax, ds
	shl	eax, 4
	or	DWORD PTR loaderGDTPtr[DATA16][2], eax

	; patch GDTR base field and load GDT register

	add	loaderGDTPtr.gdtrBase, eax
	lgdt	loaderGDTPtr

	mov	ebx, pm32Segment
	shl	ebx, 4
	add	DWORD PTR stdJumpToPM32[2], ebx

	mov	eax, cr0
	inc	ax
	mov	cr0, eax

	FLUSH_PREFETCH_QUEUE

	mov	ax, DATA32
	mov	ds, ax
	mov	es, ax
	mov	ss, ax
	mov	fs, ax

	mov	ax, DATA16
	mov	gs, ax

stdJumpToPM32:
	jmp32	stdLoaderPM32

stdBackToPM16:
	mov	ax, DATA16
	mov	ds, ax
	mov	es, ax
	mov	ss, ax
	mov	fs, ax
	mov	gs, ax

	mov	eax, cr0
	and	eax, NOT (CR0_PG + CR0_PE)
	mov	cr0, eax

	FLUSH_PREFETCH_QUEUE

	mov	ax, dataSegment
	mov	ds, ax
	mov	es, ax
	mov	fs, ax
	mov	gs, ax

	lss	esp, oldStackPtr

	ret

stdLoaderPM16 ENDP

;---------------------------------------------------------------------------
;	stdDebugInit
;
;	Entry:
;		None
;	Exit:
;		None
;
stdDebugInit PROC NEAR

	ret

stdDebugInit ENDP

;---------------------------------------------------------------------------
;	stdDebugDone
;
;	Entry:
;		None
;	Exit:
;		None
;
stdDebugDone PROC NEAR

	push	ax

	in	al, 60h

	pop	ax
	ret

stdDebugDone ENDP

;---------------------------------------------------------------------------
;	xmsInitDriver -	get extended memory driver entry point.
;
;	Entry:
;		None
;	Exit:
;		None
;
xmsInitDriver PROC NEAR

	push	ax
	push	bx
	push	es

	mov	ax, 4300h		; installation check
	int	2Fh			;
	cmp	al, 80h			; installed?
	jne	xmsInitDrvFail		;

	mov	ax, 4310h		; get entry point address
	int	2Fh			;
	mov	xmsDriverOfs, bx	; save offset
	mov	xmsDriverSeg, es	; save segment

	clc

xmsInitDrvExit:
	pop	es
	pop	bx
	pop	ax
	ret

xmsInitDrvFail:
	stc
	jmp	xmsInitDrvExit

xmsInitDriver ENDP

;---------------------------------------------------------------------------
;	xmsGrabMemory - Grab all possible Extended Memory
;
;	Entry:
;		None
;	Exit:
;		None
;
xmsGrabMemory PROC NEAR

	pushad

	mov	ah, XMS_LOCAL_A20_ENABLE
	call	xmsDriver

	mov	bp, 65535
	mov	di, OFFSET xmsHandleDef

xmsGrabMemNext:
	call	stdGetFirstFreeMDE
	jc	xmsGrabMemDone

	cmp	di, OFFSET xmsHandleEnd
	jnb	xmsGrabMemDone

xmsGrabMemMain:
	mov	dx, bp

	mov	ah, XMS_ALLOC_MEMORY
	call	xmsDriver
	test	ax, ax
	jnz	xmsGrabMemGood

	mov	ah, XMS_QUERY_MEMORY
	call	xmsDriver
	test	ax, ax
	jz	xmsGrabMemDone

	xchg	ax, bp
	jmp	xmsGrabMemMain

xmsGrabMemGood:
	mov	[di], dx

	movzx	eax, bp
	shl	eax, 10
	mov	[si].mdtSize, eax

	mov	ah, XMS_LOCK_MEMORY
	call	xmsDriver
	test	ax, ax
	jz	xmsGrabMemFail

	mov	WORD PTR [si].mdtBase[0], bx
	mov	WORD PTR [si].mdtBase[2], dx

	inc	di
	inc	di
	jmp	xmsGrabMemNext

xmsGrabMemDone:
	clc
	jmp	xmsGrabMemExit

xmsGrabMemFail:
	mov	ah, XMS_FREE_MEMORY
	call	xmsDriver

	xor	eax, eax
	mov	[si].mdtSize, eax

	stc

xmsGrabMemExit:
	popad
	ret

xmsGrabMemory ENDP

;---------------------------------------------------------------------------
;	xmsFreeMemory -	Release any Extended Memory we may have Reserveated.
;
;	Entry:
;		None
;	Exit:
;		None
;
xmsFreeMemory PROC NEAR

	pushad

	mov	di, OFFSET xmsHandleDef

xmsFreeMemNext:
	mov	dx, [di]
	test	dx, dx
	jz	xmsFreeMemExit

	mov	ah, XMS_UNLOCK_MEMORY
	call	xmsDriver
	mov	ah, XMS_FREE_MEMORY
	call	xmsDriver

	inc	di
	inc	di
	cmp	di, OFFSET xmsHandleEnd
	jb	xmsFreeMemNext

xmsFreeMemExit:
	mov	ah, XMS_LOCAL_A20_DISABLE
	call	xmsDriver

	popad
	ret

xmsFreeMemory ENDP

;---------------------------------------------------------------------------
;	dosGrabMemory - Grab all possible DOS memory
;
;	Entry:
;		None
;	Exit:
;		None
;
dosGrabMemory PROC NEAR

	pushad

dosGrabMemoryNext:
	call	stdGetFirstFreeMDE
	jc	dosGrabMemoryDone

	; query largest memory block

	mov	ah, 48h
	mov	bx, 0FFFFh
	int	21h

	test	bx, bx
	jz	dosGrabMemoryDone

	; attempt to Reserveate block

	mov	ah, 48h
	int	21h

	movzx	eax, ax
	shl	eax, 4

	movzx	ebx, bx
	shl	ebx, 4

	; store base address size and type

	mov	[si].mdtBase, eax
	mov	[si].mdtSize, ebx

	jmp	dosGrabMemoryNext

dosGrabMemoryDone:
	popad
	ret

dosGrabMemory ENDP

;---------------------------------------------------------------------------
;	rawConvMemory
;
;	Entry:
;		ebx = memory base
;		ecx = memory size
;	Exit:
;		None
;
rawConvMemory PROC NEAR

	call	stdGetFirstFreeMDE

	mov	[si].mdtBase, ebx
	mov	[si].mdtSize, ecx

	ret

rawConvMemory ENDP

;---------------------------------------------------------------------------
;	raw8800Memory
;
;	Entry:
;		None
;	Exit:
;		None
;
raw8800Memory PROC NEAR

	push	eax

	mov	ah, 88h
	int	15h
	jc	raw8800MemoryExit

	cmp	ax, 1
	jb	raw8800MemoryExit

	call	stdGetFirstFreeMDE

	movzx	eax, ax
	shl	eax, 10
	mov	[si].mdtBase, _1MEG
	mov	[si].mdtSize, eax

	clc

raw8800MemoryExit:
	pop	eax
	ret

raw8800Memory ENDP

;---------------------------------------------------------------------------
;	rawE801Memory
;
;	Entry:
;		None
;	Exit:
;		None
;
rawE801Memory PROC NEAR

	pushad

	mov	ax, 0E801h
	int	15h
	jc	rawE801MemoryExit

	call	stdGetFirstFreeMDE
	jc	rawE801MemoryExit

	; store "below 16MB" memory info

	movzx	eax, cx
	shl	eax, 10
	mov	[si].mdtBase, _1MEG
	mov	[si].mdtSize, eax

	call	stdGetFirstFreeMDE
	jc	rawE801MemoryExit

	; store "above 16MB" memory info

	movzx	eax, dx
	shl	eax, 16
	mov	[si].mdtBase, _1MEG * 16
	mov	[si].mdtSize, eax

	clc

rawE801MemoryExit:
	popad
	ret

rawE801Memory ENDP

;---------------------------------------------------------------------------
;	rawE820Memory
;
;	Entry:
;		None
;	Exit:
;		None
;
rawE820Memory PROC NEAR

	pushad
	push	es

	sub	sp, SIZE MEM_SMAP_ENTRY
	mov	bp, sp

	; es:[di] -> ss:[sp]

	push	ss
	pop	es

	mov	di, sp

	xor	esi, esi
	xor	ebx, ebx

rawE820MemoryNext:
	mov	eax, 0000E820h
	mov	ecx, SIZE MEM_SMAP_ENTRY
	mov	edx, "SMAP"
	int	15h
	jc	rawE820MemoryExit

	; is this available memory?

	cmp	[bp].memSmapType, MEM_SMAP_FREE
	jne	rawE820MemoryMore

	; make sure it is above the first 640KB

	cmp	[bp].memSmapBaseLo, 0A0000h
	jb	rawE820MemoryMore

	; do not include > 4GB stuff

	cmp	[bp].memSmapBaseHi, NULL
	jne	rawE820MemoryMore

	cmp	[bp].memSmapSizeHi, NULL
	jne	rawE820MemoryMore

rawE820MemorySave:
	call	stdGetFirstFreeMDE
	jc	rawE820MemoryExit

	; store 32-bit base address

	mov	eax, [bp].memSmapBaseLo
	mov	[si].mdtBase, eax

	; store 32-bit size

	mov	eax, [bp].memSmapSizeLo
	mov	[si].mdtSize, eax

rawE820MemoryMore:
	test	ebx, ebx
	jnz	rawE820MemoryNext

	dec	ebx

rawE820MemoryExit:
	add	sp, SIZE MEM_SMAP_ENTRY
	cmp	ebx, 1
	pop	es
	popad
	ret

rawE820Memory ENDP

;---------------------------------------------------------------------------
;	rawFindMemory - Find all possible raw BIOS reported memory
;
;	Entry:
;		None
;	Exit:
;		None
;
rawFindMemory PROC NEAR

	call	rawE820Memory
	jnc	rawFindMemoryExit

	call	rawE801Memory
	jnc	rawFindMemoryExit

	call	raw8800Memory
	jnc	rawFindMemoryExit

	;) ooough, no extended memory at all???

rawFindMemoryExit:
	ret

rawFindMemory ENDP

;---------------------------------------------------------------------------
;	kbcWaitEmpty
;
;	Entry:
;	      	None
;	Exit:
;		None
;
kbcWaitEmpty PROC NEAR

	push	ax
	push	ecx

	mov	ecx, 00200000h

kbcWaitEmptyNext:
	in	al, 64h
	test	al, 00000001b
	jz	kbcWaitEmptySkip

	in	al, 60h

kbcWaitEmptySkip:
	test	al, 00000010b
	loopdnz	kbcWaitEmptyNext

	pop	ecx
	pop	ax
	ret

kbcWaitEmpty ENDP

;---------------------------------------------------------------------------
;	kbcEnableA20
;
;	Entry:
;	      	None
;	Exit:
;		None
;
kbcEnableA20 PROC NEAR

	push	ax

	call	kbcWaitEmpty		;
	mov	al, 0D1h		; write output port
	out	64h, al			;

	call	kbcWaitEmpty		;
	mov	al, 0DFh		; set A20 enable bit
	out	60h, al			;

	call	kbcWaitEmpty		;
	mov	al, 0FFh		; end of sequence command
	out	64h, al			;

	pop	ax
	ret

kbcEnableA20 ENDP

;---------------------------------------------------------------------------
;	kbcDisableA20
;
;	Entry:
;	      	None
;	Exit:
;		None
;
kbcDisableA20 PROC NEAR

	push	ax

	call	kbcWaitEmpty		;
	mov	al, 0D1h		; write output port
	out	64h, al			;

	call	kbcWaitEmpty		;
	mov	al, 0DDh		; clear A20 enable bit
	out	60h, al			;

	call	kbcWaitEmpty		;
	mov	al, 0FFh		; end of sequence command
	out	64h, al			;

	pop	ax
	ret

kbcDisableA20 ENDP

;---------------------------------------------------------------------------
;	ps2EnableA20
;
;	Entry:
;	      	None
;	Exit:
;		None
;
ps2EnableA20 PROC NEAR

	push	ax
	in	al, 92h			;
	or	al, 00000010b		; PS/2 fast gate A20 enable
	out	92h, al			;
	pop	ax
	ret

ps2EnableA20 ENDP

;---------------------------------------------------------------------------
;	ps2DisableA20
;
;	Entry:
;	      	None
;	Exit:
;		None
;
ps2DisableA20 PROC NEAR

	push	ax
	in	al, 92h			;
	and	al, 11111101b		; PS/2 fast gate A20 disable
	out	92h, al			;
	pop	ax
	ret

ps2DisableA20 ENDP

;---------------------------------------------------------------------------
;	stdGetFirstMDE
;
;	Entry:
;	      	None
;	Exit:
;		None
;
stdGetFirstMDE PROC NEAR

	mov	si, OFFSET loaderMDTDef
	ret

stdGetFirstMDE ENDP

;---------------------------------------------------------------------------
;	stdGetNextMDE
;
;	Entry:
;	      	None
;	Exit:
;		None
;
stdGetNextMDE PROC NEAR

	add	si, SIZE MDT
	cmp	si, OFFSET loaderMDTEnd
	cmc
	ret

stdGetNextMDE ENDP

;---------------------------------------------------------------------------
;	stdGetFirstFreeMDE
;
;	Entry:
;	      	None
;	Exit:
;		None
;
stdGetFirstFreeMDE PROC NEAR

	call	stdGetFirstMDE

stdGetFirstFreeMDEFind:
	cmp	[si].mdtSize, NULL
	je	stdGetFirstFreeMDEExit

	call	stdGetNextMDE
	jnc	stdGetFirstFreeMDEFind

stdGetFirstFreeMDEExit:
	ret

stdGetFirstFreeMDE ENDP

;---------------------------------------------------------------------------
;	stdGetNextFreeMDE
;
;	Entry:
;	      	None
;	Exit:
;		None
;
stdGetNextFreeMDE PROC NEAR

stdGetNextFreeMDEFind:
	call	stdGetNextMDE
	jc	stdGetNextFreeMDEExit

	cmp	[si].mdtSize, NULL
	jne	stdGetNextFreeMDEFind

stdGetNextFreeMDEExit:
	ret

stdGetNextFreeMDE ENDP

;---------------------------------------------------------------------------
;	stdGetFirstUsedMDE
;
;	Entry:
;	      	None
;	Exit:
;		None
;
stdGetFirstUsedMDE PROC NEAR

	call	stdGetFirstMDE

stdGetFirstUsedMDEFind:
	cmp	[si].mdtSize, NULL
	jne	stdGetFirstUsedMDEExit

	call	stdGetNextMDE
	jnc	stdGetFirstUsedMDEFind

stdGetFirstUsedMDEExit:
	ret

stdGetFirstUsedMDE ENDP

;---------------------------------------------------------------------------
;	stdGetNextUsedMDE
;
;	Entry:
;	      	None
;	Exit:
;		None
;
stdGetNextUsedMDE PROC NEAR

stdGetNextUsedMDEFind:
	call	stdGetNextMDE
	jc	stdGetNextUsedMDEExit

	cmp	[si].mdtSize, NULL
	je	stdGetNextUsedMDEFind

stdGetNextUsedMDEExit:
	ret

stdGetNextUsedMDE ENDP

;---------------------------------------------------------------------------
;	stdPrintChar
;
;	Entry:
;		al = character
;	Exit:
;		None
;
stdPrintChar PROC NEAR

	push	ax
	push	bx

	mov	ah, 0Eh
	mov	bh, 00h
	int	10h

	pop	bx
	pop	ax
	ret

stdPrintChar ENDP

;---------------------------------------------------------------------------
;	stdPrintCRLF
;
;	Entry:
;		None
;	Exit:
;		None
;
stdPrintCRLF PROC NEAR

	push	ax
	mov	al, 0Dh
	call	stdPrintChar
	mov	al, 0Ah
	call	stdPrintChar
	pop	ax
	ret

stdPrintCRLF ENDP

;---------------------------------------------------------------------------
;	stdPrintMsgZ
;
;	Entry:
;		si = pointer to null-terminated message
;	Exit:
;		None
;
stdPrintMsgZ PROC NEAR

	push	ax
	push	si

stdPrintMsgZNext:
	lodsb
	test	al, al
	jz	stdPrintMsgZDone

	call	stdPrintChar
	jmp	stdPrintMsgZNext

stdPrintMsgZDone:
	pop	si
	pop	ax
	ret

stdPrintMsgZ ENDP

codeSegment ENDS

;---------------------------------------------------------------------------
; 32-bit Protected Mode
;---------------------------------------------------------------------------

pm32Segment SEGMENT USE32 'pm32'

	ASSUME	cs:pm32Segment, gs:dataSegment, ds:NOTHING

;---------------------------------------------------------------------------
;	pm32FindObject
;
;	Entry:
;		esi = pointer to ASCIIZ "object name" string
;	Exit:
;		ebx = object base
;		ecx = object size
;
pm32FindObject PROC NEAR

	push	eax
	push	esi
	push	edi
	push	edi
	push	edi

	mov	edi, esp		; edi -> name buffer

	xor	ecx, ecx
	mov	cl, 8

	lodsb				;
	cmp	al, "*"			; embedded object?
	jne	pm32FindObjectFail	;

pm32FindObjectCopy:
	lodsb				;
	test	al, al			; end of name?
	jz	pm32FindObjectFill	;

	cmp	al, "a"			; below lowercase range?
	jb	pm32FindObjectSkip	;
	cmp	al, "z"			; above lowercase range?
	ja	pm32FindObjectSkip	;
	sub	al, " "			; convert to uppercase

pm32FindObjectSkip:
	stosb				;
	dec	ecx			; store character
	jnz   	pm32FindObjectCopy	;

pm32FindObjectFill:
	rep	stosb			; possibly pad with zeroes

	mov	ebx, ourLibraryBase	; ebx -> library
	test	ebx, ebx
	jz	pm32FindObjectFail	

	mov	esi, ebx
	add	esi, [ebx].mzExeLibHeader

	cmp	[esi].mzLibSignature, MZ_LIB_MAGIC
	jne	pm32FindObjectFail

	mov	ecx, [esi].mzLibCount	; any directory entries?
	jecxz	pm32FindObjectFail	;

	; point to first entry and start searching

	add	esi, SIZE MZ_LIB

pm32FindObjectFind:
	mov	edi, esp		; edi -> "Name"

	push	ecx
	push	esi

	mov	cl, 2
	rep	cmpsd

	pop	esi
	pop	ecx
	je	pm32FindObjectDone

	add	esi, SIZE MZ_OBJ	; esi -> next entry
	dec	ecx			;
	jnz	pm32FindObjectFind	; more?

pm32FindObjectFail:
	stc				; failure, not found
	jmp	pm32FindObjectExit	;

	; object was found, return objBase & objSize

pm32FindObjectDone:
	add	ebx, [esi].mzObjBase
	mov	ecx, [esi].mzObjSize

	clc

pm32FindObjectExit:
	pop	edi
	pop	edi
	pop	edi
	pop	esi
	pop	eax
	ret

pm32FindObject ENDP

;---------------------------------------------------------------------------
;	pm32PeValidate
;
;	Entry:
;		ebx = pointer to "PE" image
;
;	Exit (Success: CF==0):
;
;		ecx = image size
;
pm32PeValidate PROC NEAR

	push	ebx

	cmp	[ebx].mzSignature, MZ_EXE_MAGIC
	jne	pm32PeValidateFail

	add	ebx, [ebx].mzNewExeHeader

	cmp	[ebx].peSignature, PE_EXE_MAGIC
	jne	pm32PeValidateFail

	mov	ecx, [ebx].peImageSize

	clc

pm32PeValidateExit:
	pop	ebx
	ret

pm32PeValidateFail:
	stc
	jmp	pm32PeValidateExit

pm32PeValidate ENDP

;---------------------------------------------------------------------------
;	pm32PeRelocate
;
;	Entry:
;		ebx = pointer to "PE" image
;	Exit:
;		None
;
pm32PeRelocate PROC NEAR

	pushad

	mov	edi, ebx
	add	edi, [edi].mzNewExeHeader

	mov	edx, ebx
	sub	edx, [edi].peImageBase
	mov	[edi].peImageBase, ebx

	mov	esi, [edi].peFixupRVA
	mov	ecx, [edi].peFixupSize
	jecxz	pm32PeRelocateDone

pm32PeRelocateMore:
	mov	ebp, [ebx][esi].peFixPageRVA
	mov	ecx, [ebx][esi].peFixBlockSize
	jecxz	pm32PeRelocateDone

	sub	ecx, 8
	add	esi, 8

pm32PeRelocateNext:
	mov	eax, [ebx][esi]
	mov	edi, eax

	and	ah, 11110000b
	cmp	ah, 00110000b
	jne	pm32PeRelocateSkip

	and	edi, 00000FFFh
	add	edi, ebp
	add	[ebx][edi], edx

pm32PeRelocateSkip:
	inc	esi
	inc	esi
	dec	ecx
	dec	ecx
	jnz	pm32PeRelocateNext
	jmp	pm32PeRelocateMore

pm32PeRelocateDone:
	clc
	popad
	ret

pm32PeRelocate ENDP

;---------------------------------------------------------------------------
;	pm32PeLoadModule
;
;	Entry:
;		ebx = pointer to "PE" image
;		edi = pointer to target memory block
;	Exit:
;		eax = entry point RVA
;
pm32PeLoadModule PROC NEAR

	push	ebx
	push	ecx
	push	esi

	; eax -> "PE" header, ebx/esi -> "MZ" header

	mov	eax, ebx
	mov	esi, ebx
	add	eax, [eax].mzNewExeHeader

	; clear memory block

	push	eax
	push	edi

	mov	ecx, [eax].peImageSize
	shr	ecx, 2
	xor	eax, eax
	rep	stosd

	pop	edi
	pop	eax

	; copy image header

	push	esi
	push	edi

	mov	ecx, [eax].peHeaderSize
	shr	ecx, 2
	rep	movsd

	pop	edi
	pop	esi

	movzx	ecx, [eax].peObjectCount
	movzx	ebx, [eax].peNTHeaderSize

	add	ebx, peOptionalStart
	add	ebx, eax

	; copy all sections

pm32PeLoadModuleCopy:
	push	ecx
	push	esi
	push	edi

	add	esi, [ebx].peObjPhysicalOfs
	add	edi, [ebx].peObjRVA
	mov	ecx, [ebx].peObjPhysicalSize
	shr	ecx, 2
	rep	movsd

	pop	edi
	pop	esi
	pop	ecx

	add	ebx, SIZE PE_OBJ
	dec	ecx
	jnz	pm32PeLoadModuleCopy

	; return entry point RVA

	mov	eax, [eax].peEntryPoint

	pop	esi
	pop	ecx
	pop	ebx
	ret

pm32PeLoadModule ENDP

;---------------------------------------------------------------------------
;	pm32MmZap
;
;	Entry:
;		eax = memory base (page address)
;		ecx = memory size
;	Exit:
;	     	None
;
pm32MmZap PROC ENAR

	pushfd
	push	eax
	push	ecx
	push	edi

	mov	edi, eax
	xor	eax, eax
	shr	ecx, 2
	rep	stosd

	pop	edi
	pop	ecx
	pop	eax
	popfd
	ret

pm32MmZap ENDP

;---------------------------------------------------------------------------
;	pm32MmMap
;
;	Entry:
;		ebx = PTE
;		eax = memory base (physical)
;		ecx = memory size (bytes)
;		 dl = 8-bits page flags
;	Exit:
;	     	ebx = updated beyond last PTE
;		eax = updated beyond last page address
;
pm32MmMap PROC NEAR

	jecxz	pm32MmMapExit

	push	ecx

	add	ecx, 000000FFFh		; round-up
	shr	ecx, 12			;

pm32MmMapNext:
	mov	[ebx], eax		; save page address flags
	mov	[ebx], dl		; save page flags
	add	eax, _PAGE		; next 4KB page
	add	ebx, 4			; next PTE
	dec	ecx			; more?
	jnz	pm32MmMapNext		;

	mov	ecx, cr3		; flush TLBs
	mov	cr3, ecx		;

	pop	ecx

pm32MmMapExit:
	ret

pm32MmMap ENDP

;---------------------------------------------------------------------------
;	pm32MmGetMDE
;
;	Entry:
;	      	esi = MDE pointer (NULL or intermediate)
;	Exit:
;		esi = MDE pointer (next or first)
;
pm32MmGetMDE PROC NEAR

	push	eax

	LOAD32	eax, loaderMDTEnd

	test	esi, esi
	jnz	pm32MmGetMDEMain

	LOAD32	esi, loaderMDTDef

	clc
	jmp	pm32MmGetMDEExit

pm32MmGetMDEMain:
	add	esi, SIZE MDT
	cmp	esi, eax
	cmc

pm32MmGetMDEExit:
	pop	eax
	ret

pm32MmGetMDE ENDP

;---------------------------------------------------------------------------
;	pm32MmReserve - Reserves memory above 1MB
;
;	Entry:
;		ecx = memory size needed
;
;	Exit (Success: CF==0):
;
;		eax = memory base
;
pm32MmReserve PROC NEAR

	push	ebx
	push	ecx
	push	edx
	push	esi
	push	edi

	xor	esi, esi		; get first MDE
	call	pm32MmGetMDE		;

	PAGE_ALIGN ecx

pm32MmReserveFind:
	mov	ebx, [esi].mdtBase	;
	cmp	ebx, _1MEG		; above 1MB?
	jb	pm32MmReserveSkip	;

	mov	eax, ebx		; get page aligned address
	PAGE_ALIGN eax			;

	mov	edx, eax		; get difference in bases
	sub	edx, ebx		;

	mov	edi, [esi].mdtSize	; get memory block size
	sub	edi, edx		;

	cmp	edi, ecx		; enough?
	jb	pm32MmReserveSkip	;

	add	ecx, edx

	add	[esi].mdtBase, ecx	; reserve memory
	sub	[esi].mdtSize, ecx	;

	clc				; success
	jmp	pm32MmReserveExit	;

pm32MmReserveSkip:
	call	pm32MmGetMDE		; more?
	jnc	pm32MmReserveFind	;

pm32MmReserveExit:
	pop	edi
	pop	esi
	pop	edx
	pop	ecx
	pop	ebx
	ret

pm32MmReserve ENDP

;---------------------------------------------------------------------------
;	stdLoaderPM32
;
;	Entry:
;		
;	       ebx = linear base of "pm32Segment"
;
;		cs = CODE32 (32-bit Flat R/E Code Selector)
;
;		ds = DATA32 (32-bit Flat R/W Data Selector)
;		es = DATA32 (32-bit Flat R/W Data Selector)
;
;		gs = DATA16 (16-bit R/W Data Selector for "dataSegment")
;
;	    ss:esp = 2KB stack
;
;	Exit:
;		None
;
;
;	KERNEL32 will be called with the following parameters:
;
;	Registers:
;
;		ebx = our memory block base -- "module handle"
;		ecx = our memory block size
;		ebp = pointer to arguments on the stack
;
;	Stack (ss:esp):
;
;		DWORD = return address
;
;	Stack Frame (ss:ebp):
;
;		DWORD	Base of EXE image
;		DWORD	Size of EXE image
;		DWORD	Base of reclaim area
;		DWORD	Size of reclaim area
;		DWORD	Base of MDT
;		DWORD	Size of MDT (number of entries)
;		DWORD	Base of PSP
;		DWORD	Base of DOS Environment
;
stdLoaderPM32 PROC FAR

	;-------------------------------------------------------------
	; push KERNEL32 startup arguments
	;-------------------------------------------------------------

	movzx	eax, ourEnvSeg		;
	shl	eax, 4			; Flat -> ourEnvSeg
	push	eax			;

	movzx	eax, ourPSPSeg		;
	shl	eax, 4			; Flat -> ourPSPSeg
	push	eax			;
	
	xor	esi, esi		; get first MDE
	call	pm32MmGetMDE		;

	push	LARGE (loaderMDTEnd-loaderMDTDef)/SIZE MDT
	push	esi

	push	ourReclaimSize		; push "Reclaim" size
	push	ourReclaimBase		; push "Reclaim" base address

	push	ourLibrarySize		; push "Library" size
	push	ourLibraryBase		; push "Library" base address

	mov	ebp, esp		; ebp -> stack frame

	;-------------------------------------------------------------
	; create paging structures
	;-------------------------------------------------------------

	mov	ecx, _PAGE * 3		; 12KB

	call	pm32MmReserve		; get memory for PD and PTs
	jc	stdExitToPM16		;

	call	pm32MmZap		; zap memory block

	xchg	eax, edi		; edi -> PD

	lea	ebx, [edi][_PAGE]	; ebx -> PT

	mov	dl, PAGE_RING3		; U/W/P

	mov	[edi], ebx		; PDE "00000000" -> PT (Lower 4MB)
	mov	[edi], dl		;

	xor	eax, eax		; map from address zero
	mov	ecx, _PAGE * 1024	; map 1024 pages
	call	pm32MmMap		; identity map the first 4MB

	mov	dl, PAGE_RING0		; S/W/P

	mov	[edi][0FFCh], edi	; PDE "FFFFF000" -> PD/PT
	mov	[edi][0FFCh], dl	;

	mov	[edi][0C00h], ebx	; PDE "C0000000" -> PT (Kernel)
	mov	[edi][0C00h], dl	;

	mov	cr3, edi		; cr3 -> PD

	mov	eax, cr0		;
	or	eax, CR0_PG		; turn paging ON!
	mov	cr0, eax		;

	;-------------------------------------------------------------
	; search for and reserve memory for kernel
	;-------------------------------------------------------------

	LOAD32	esi, szKernel32

	call	pm32FindObject		; search for KERNEL32 object
	jc	stdExitToPM16		;

	call	pm32PeValidate		; is this a valid EXE?
	jc	stdExitToPM16		;

	call	pm32MmReserve		; reserve memory for image
	jc	stdExitToPM16		;

	;-------------------------------------------------------------
	; map KERNEL32 image starting at linear 3GB
	;-------------------------------------------------------------

	push	ebx

	mov	ebx, 0FFF00000h		; PTE "C0000000" -> KERNEL32
	call	pm32MmMap		; map N pages

	pop	ebx

	;-------------------------------------------------------------
	; load, relocate and execute kernel
	;-------------------------------------------------------------

	mov	edi, 0C0000000h		; load module at 3GB
	call	pm32PeLoadModule	;

	mov	ebx, edi		; ebx -> PE

	call	pm32PeRelocate		; process image fixups

	add	eax, ebx		; eax -> entry point

	call	eax			; call KERNEL32.mainEntryPoint

stdExitToPM16:
	jmp16	stdBackToPM16

stdLoaderPM32 ENDP

pm32Segment ENDS

;---------------------------------------------------------------------------
; Initialized Data
;---------------------------------------------------------------------------

dataSegment SEGMENT USE16 'data'

; GDT -- Global Descriptor Table

loaderGDTPtr	LABEL	FWORD
	DW	loaderGDTEnd-loaderGDTPtr-1, loaderGDTPtr, ?, ?
loaderGDTDef	LABEL	DWORD
	GDT	<0FFFFh, 0, 0, 09Bh, 000h, 0>	; 0x08 16-bit code
	GDT	<0FFFFh, 0, 0, 093h, 000h, 0>	; 0x10 16-bit data
	GDT	<0FFFFh, 0, 0, 09Bh, 0CFh, 0>	; 0x18 32-bit code (Ring0)
	GDT	<0FFFFh, 0, 0, 093h, 0CFh, 0>	; 0x20 32-bit data (Ring0)
	GDT	<0FFFFh, 0, 0, 0FBh, 0CFh, 0>	; 0x28 32-bit code (Ring3)
	GDT	<0FFFFh, 0, 0, 0F3h, 0CFh, 0>	; 0x30 32-bit data (Ring3)
	GDT	<0FFFFh, 0, 0, 089h, 000h, 0>	; 0x38 TSS
loaderGDTEnd	LABEL	DWORD

; MDT -- Memory Descriptor Table

loaderMDTDef	LABEL	DWORD
	MDT	8 DUP (<>)
loaderMDTEnd	LABEL	DWORD

; XMS handle table

xmsHandleDef	LABEL	WORD
	DW	8 DUP (?)
xmsHandleEnd	LABEL	WORD

; XMS driver entry point

xmsDriver	LABEL	DWORD
xmsDriverOfs	DW	?
xmsDriverSeg	DW	?

; Misc

ourPSPSeg	DW	?
ourEnvSeg	DW	?

ourLibraryBase	DD	?
ourLibrarySize	DD	?

ourReclaimBase	DD	?
ourReclaimSize	DD	?

oldStackPtr	LABEL	FWORD
oldStackOfs	DD	?
oldStackSeg	DD	?

; Messages

msgLoader	DB	"Operating System Loader v1.1 [RM] ", BUILD, CRLF, COPYRIGHT, CRLF, NULL
msgReboot	DB	"System will now reboot, press any key to continue...", NULL
msgErrV86	DB	CRLF, "- System is already in V86/Protected Mode!", CRLF, BELL, NULL
msgErrLib	DB	CRLF, "- Unable to load boot library into memory!", CRLF, BELL, NULL

szKernel32	DB	"*kernel32",	NULL

dataSegment ENDS

;---------------------------------------------------------------------------
; Stack Segment
;---------------------------------------------------------------------------

stackSegment SEGMENT USE16 STACK 'stack'
	DB	2048 DUP (?)
stackSegment ENDS

;---------------------------------------------------------------------------
; End Marker
;---------------------------------------------------------------------------

lastSegment SEGMENT USE16 'last'

lastSegment ENDS

	END	mainEntryPoint
