;---------------------------------------------------------------------------
;
;	      Mini Driver for standard AT 101/102-key keyboards
;
;     Copyright (c) 1999-2000 by Thomas Kjoernes, all rights reserved.
;
;---------------------------------------------------------------------------

	INCLUDE std.inc
	INCLUDE stdgroup.inc
	INCLUDE kernel.api
	INCLUDE keyboard.api
	INCLUDE	isa.api
	INCLUDE	keyboard.inc

	.386p

	; Publics

	PUBLIC	kbGetFlags
	PUBLIC	kbGetLayout
	PUBLIC	kbSetLayout
	PUBLIC	kbEnable
	PUBLIC	kbDisable
	PUBLIC	kbUpdateLEDs
	PUBLIC	kbSetRateDelay

codeSegment SEGMENT

	; Externs

;---------------------------------------------------------------------------
;	mainEntryPoint - MINIKBD.SYS entry point
;
;	Entry (Stack):
;
;		DWORD = return address
;
;	Entry (Registers):
;
;		ebx = our memory block base -- "module handle"
;		ecx = our memory block size
;
DefProc mainEntryPoint

	; register keyboard services

	mov	esi, OFFSET apiServiceTbl
	APIcall	KERNEL_ApiRegister

	; install keyboard handler and unmask IRQ
	
	mov	al, IRQ1
	mov	ebx, OFFSET kbEventHandler
	APIcall	KERNEL_IrqSetHandlerPM

	mov	kbOldHandler, ebx
	
	APIcall	KERNEL_IrqUnMask

	clc
	ret

EndProc mainEntryPoint

;---------------------------------------------------------------------------
;	kbGetVersion - 
;
;	Entry:
;		None
;	Exit:
;		eax = version ID
;
DefProc kbGetVersion

	mov	eax, 00000100h
	clc
	ret

EndProc kbGetVersion

;---------------------------------------------------------------------------
;	kbGetFlags
;
;	Entry:
;	      	None
;	Exit:
;	     	al = Keyboard Flags
;		bl = Keyboard Shift
;		cl = Keyboard LEDs
;
DefProc kbGetFlags

	mov	al, kbFlags
	mov	bl, kbShift
	mov	cl, kbState
	ret

EndProc kbGetFlags

;---------------------------------------------------------------------------
;	kbGetLayout
;
;	Entry:
;	      	None
;	Exit:
;	     	ebx = pointer to KB_LAYOUT structure
;
DefProc kbGetLayout

	mov	ebx, kbLayout
	ret

EndProc kbGetLayout

;---------------------------------------------------------------------------
;	kbSetLayout
;
;	Entry:
;	     	ebx = pointer to new KB_LAYOUT structure
;	Exit:
;	     	ebx = pointer to old KB_LAYOUT structure
;
DefProc kbSetLayout

	xchg	ebx, kbLayout
	ret

EndProc kbSetLayout

;---------------------------------------------------------------------------
;	kbEnable
;
;	Entry:
;	      	None
;	Exit:
;	     	al = AEh
;
DefProc kbEnable

	mov	al, 0AEh
	APIcall	ISA_KBC_SendCommand
	ret

EndProc kbEnable

;---------------------------------------------------------------------------
;	kbDisable
;
;	Entry:
;	      	None
;	Exit:
;	     	al = ADh
;
DefProc kbDisable

	mov	al, 0ADh
	APIcall	ISA_KBC_SendCommand
	ret

EndProc kbDisable

;---------------------------------------------------------------------------
;	kbUpdateLEDs
;
;	Entry:
;	      	al = new LED state
;	Exit:
;	     	None
;
DefProc kbUpdateLEDs

	push	eax

	and	al, 00000111b
	mov	ah, al

	mov	al, 0EDh
	APIcall	ISA_KBC_SendToKbd
	APIcall	ISA_KBC_ReceiveFromKbd

	mov	al, ah
	APIcall	ISA_KBC_SendToKbd
	APIcall	ISA_KBC_ReceiveFromKbd

	pop	eax
	ret

EndProc kbUpdateLEDs

;---------------------------------------------------------------------------
;	kbSetRateDelay
;
;	Entry:
;	      	al = RateDelay rate/delay
;	Exit:
;	     	None
;
DefProc kbSetRateDelay

	push	eax

	and	al, 01111111b
	mov	ah, al

	mov	al, 0F3h
	APIcall	ISA_KBC_SendToKbd
	APIcall	ISA_KBC_ReceiveFromKbd

	mov	al, ah
	APIcall	ISA_KBC_SendToKbd
	APIcall	ISA_KBC_ReceiveFromKbd

	pop	eax
	ret

EndProc kbSetRateDelay

;---------------------------------------------------------------------------
;	kbCheckFlags
;
;	Entry:
;	      	al = scancode
;	Exit:
;	     	ah = altered
;
;		CF==0 if AL was FF/FE/FA/E0/E1
;
DefProc kbCheckFlags

	mov	ah, KB_FLAGS_ERROR
	cmp	al, 0FFh
	je	kbCheckFlagsDone

	mov	ah, KB_FLAGS_RESEND
	cmp	al, 0FEh
	je	kbCheckFlagsDone

	mov	ah, KB_FLAGS_ACK
	cmp	al, 0FAh
	je	kbCheckFlagsDone

	mov	ah, KB_FLAGS_E0
	cmp	al, 0E0h
	je	kbCheckFlagsDone

	mov	ah, KB_FLAGS_E1_B
	cmp	al, 0E1h
	je	kbCheckFlagsDone

kbCheckFlagsFail:
	stc
	ret

kbCheckFlagsDone:
	or	kbFlags, ah
	clc
	ret

EndProc kbCheckFlags

;---------------------------------------------------------------------------
;	kbCheckShiftKeys
;
;	Entry:
;	      	al = scancode
;	Exit:
;	     	ah = modified
;		bl = modified
;
;		CF==0 if AL was a valid shift state key press/relase
;
DefProc kbCheckShiftKeys

	mov	bl, al
	and	bl, 01111111b

	test	kbFlags, KB_FLAGS_E0
	jnz	kbCheckShiftKeysExtended

	mov	ah, KB_SHIFT_LEFT_SHIFT
	cmp	bl, KB_SCAN_LEFT_SHIFT
	je	kbCheckShiftKeysDone

	mov	ah, KB_SHIFT_RIGHT_SHIFT
	cmp	bl, KB_SCAN_RIGHT_SHIFT
	je	kbCheckShiftKeysDone

	mov	ah, KB_SHIFT_LEFT_CTRL
	cmp	bl, KB_SCAN_CTRL
	je	kbCheckShiftKeysDone

	mov	ah, KB_SHIFT_LEFT_ALT
	cmp	bl, KB_SCAN_ALT
	je	kbCheckShiftKeysDone

kbCheckShiftKeysFail:
	stc
	ret

kbCheckShiftKeysExtended:

	mov	ah, KB_SHIFT_NONE

	cmp	bl, KB_SCAN_LEFT_SHIFT
	je	kbCheckShiftKeysExtendedDone

	cmp	bl, KB_SCAN_RIGHT_SHIFT
	je	kbCheckShiftKeysExtendedDone

	mov	ah, KB_SHIFT_RIGHT_ALT
	cmp	bl, KB_SCAN_ALT
	je	kbCheckShiftKeysExtendedDone

	mov	ah, KB_SHIFT_RIGHT_CTRL
	cmp	bl, KB_SCAN_CTRL
	je	kbCheckShiftKeysExtendedDone

	mov	ah, KB_SHIFT_RIGHT_WIN
	cmp	bl, KB_SCAN_RIGHT_WIN
	je	kbCheckShiftKeysExtendedDone

	mov	ah, KB_SHIFT_LEFT_WIN
	cmp	bl, KB_SCAN_LEFT_WIN
	jne	kbCheckShiftKeysFail

kbCheckShiftKeysExtendedDone:
	and	kbFlags, NOT KB_FLAGS_E0

kbCheckShiftKeysDone:
	test	al, al
	js	kbCheckShiftKeysClear

kbCheckShiftKeysSet:
	or	kbShift, ah
	clc
	ret

kbCheckShiftKeysClear:
	not	ah
	and	kbShift, ah
	clc
	ret

EndProc kbCheckShiftKeys

;---------------------------------------------------------------------------
;	kbCheckLEDKeys
;
;	Entry:
;	      	al = scancode
;	Exit:
;		CF==0 if AL was a valid LED state key press/relase
;
DefProc kbCheckLEDKeys

	test	kbShift, KB_SHIFT_CTRL
	jnz	kbCheckLEDKeysFail

	cmp	al, KB_SCAN_SCROLL_LOCK
	je	kbCheckLEDKeysScrollDown
	cmp	al, KB_SCAN_SCROLL_LOCK + KB_SCAN_RELEASE
	je	kbCheckLEDKeysScrollUp

	cmp	al, KB_SCAN_NUM_LOCK
	je	kbCheckLEDKeysNumDown
	cmp	al, KB_SCAN_NUM_LOCK + KB_SCAN_RELEASE
	je	kbCheckLEDKeysNumUp

	cmp	al, KB_SCAN_CAPS_LOCK
	je	kbCheckLEDKeysCapsDown
	cmp	al, KB_SCAN_CAPS_LOCK + KB_SCAN_RELEASE
	je	kbCheckLEDKeysCapsUp

kbCheckLEDKeysFail:
	stc
	ret

kbCheckLEDKeysUpdate:
	mov	al, kbState
	call	kbUpdateLEDs

kbCheckLEDKeysDone:
	clc
	ret

	; Scroll-Lock down

kbCheckLEDKeysScrollDown:
	test	kbState, KB_STATE_SCROLL_DOWN
	jnz	kbCheckLEDKeysDone

	or	kbState, KB_STATE_SCROLL_DOWN
	xor	kbState, KB_STATE_SCROLL
	jmp	kbCheckLEDKeysUpdate

	; Scroll-Lock up

kbCheckLEDKeysScrollUp:
	and	kbState, NOT KB_STATE_SCROLL_DOWN
	jmp	kbCheckLEDKeysDone

	; Num-Lock down

kbCheckLEDKeysNumDown:
	test	kbState, KB_STATE_NUM_DOWN
	jnz	kbCheckLEDKeysDone

	or	kbState, KB_STATE_NUM_DOWN
	xor	kbState, KB_STATE_NUM
	jmp	kbCheckLEDKeysUpdate

	; Num-Lock up

kbCheckLEDKeysNumUp:
	and	kbState, NOT KB_STATE_NUM_DOWN
	jmp	kbCheckLEDKeysDone

	; Caps-Lock down

kbCheckLEDKeysCapsDown:
	test	kbState, KB_STATE_CAPS_DOWN
	jnz	kbCheckLEDKeysDone

	or	kbState, KB_STATE_CAPS_DOWN
	xor	kbState, KB_STATE_CAPS
	jmp	kbCheckLEDKeysUpdate

	; Caps-Lock up

kbCheckLEDKeysCapsUp:
	and	kbState, NOT KB_STATE_CAPS_DOWN
	jmp	kbCheckLEDKeysDone

EndProc kbCheckLEDKeys

;---------------------------------------------------------------------------
;	kbCheckReboot
;
;	Entry:
;	      	al = scancode
;	Exit:
;		CF==0 if CTRL+ALT+DEL was pressed
;
DefProc kbCheckReboot

	cmp	al, KB_SCAN_DEL
	jne	kbCheckRebootDone

	test	kbShift, KB_SHIFT_ALT
	jz	kbCheckRebootDone

	test	kbShift, KB_SHIFT_CTRL
	jz	kbCheckRebootDone

	clc
	ret

kbCheckRebootDone:
	stc
	ret

EndProc kbCheckReboot

;---------------------------------------------------------------------------
;	kbCheckAltGr
;
;	Entry:
;	      	al = scancode
;	Exit:
;		CF==0 if AltGr-modified key was pressed
;
DefProc kbCheckAltGr

	test	kbShift, KB_SHIFT_RIGHT_ALT
	jnz	kbCheckAltGrMain

kbCheckAltGrFail:
	stc
	ret

kbCheckAltGrMain:
	mov	ebx, kbLayout
	mov	esi, [ebx].klTableAltGr
	test	esi, esi
	jz	kbCheckAltGrFail

kbCheckAltGrFind:
	mov	ah, [esi]

	inc	esi
	inc	esi

	cmp	ah, 0FFh
	je	kbCheckAltGrFail
	cmp	al, ah
	jne	kbCheckAltGrFind

	mov	al, [esi-1]

	clc
	ret	

EndProc kbCheckAltGr

;---------------------------------------------------------------------------
;	kbCheckExtended
;
;	Entry:
;	      	al = scancode
;	Exit:
;		CF==0 if Extended key was pressed
;
DefProc kbCheckExtended

	test	kbFlags, KB_FLAGS_E0	
	jnz	kbCheckExtendedMain

kbCheckExtendedFail:
	stc
	ret

kbCheckExtendedMain:
	mov	ebx, kbLayout
	mov	esi, [ebx].klTableExtended
	test	esi, esi
	jz	kbCheckExtendedFail

kbCheckExtendedFind:
	mov	ah, [esi]

	inc	esi
	inc	esi

	cmp	ah, 0FFh
	je	kbCheckExtendedFail
	cmp	al, ah
	jne	kbCheckExtendedFind

	mov	al, [esi-1]

	clc
	ret	

EndProc kbCheckExtended

;---------------------------------------------------------------------------
;	kbTranslate
;
;	Entry:
;		al = scancode
;
;	Exit (Success):
;
;		CF==0, eax = keycode	
;
;	Exit (Failure):
;		
;		CF==1
;
;	Modifies:
;
;		ebx, esi
;
DefProc kbTranslate

	movzx	eax, al

	mov	ebx, kbLayout
	lea	esi, [eax+eax*4]
	add	esi, [ebx].klTableMain

	mov	ah, al

	mov	bl, kbShift
	mov	bh, kbState

	; Alt?

	mov	al, [esi+4]
	test	bl, KB_SHIFT_ALT
	jnz	kbTranslateMain

	; Ctrl?

	mov	al, [esi+3]
	test	bl, KB_SHIFT_CTRL
	jnz	kbTranslateMain

	; Shift?

	mov	al, [esi+2]
	test	bl, KB_SHIFT_SHIFT
	jnz	kbTranslateMain

	; Normal?

	mov	al, [esi+1]

kbTranslateMain:
	cmp	al, 0FFh
	je	kbTranslateFail

	cmp	BYTE PTR [esi], 0
	je	kbTranslateType0
	cmp	BYTE PTR [esi], 1
	je	kbTranslateType1
	cmp	BYTE PTR [esi], 2
	je	kbTranslateType2
	cmp	BYTE PTR [esi], 3
	je	kbTranslateType3

kbTranslateFail:
	stc
	ret

	; Type 00 - Normal Key

kbTranslateType0:
	test	bl, KB_SHIFT_ALT
	jnz	kbTranslateType2

	jmp	kbTranslateDone

	; Type 01 - Normal Key - CAPS LOCK sensitive

kbTranslateType1:
	test	bl, KB_SHIFT_ALT
	jnz	kbTranslateType2

	test	bh, KB_STATE_CAPS
	jz	kbTranslateDone

	mov	al, [esi+1]

	test	bl, KB_SHIFT_SHIFT
	jnz	kbTranslateDone

	mov	al, [esi+2]
	jmp	kbTranslateDone

	; Type 02 - Function Key

kbTranslateType2:
	mov	ah, al
	xor	al, al
	jmp	kbTranslateDone

	; Type 03 - Numpad Key

kbTranslateType3:
	mov	al, [esi+1]

	test	kbFlags, KB_FLAGS_E0
	jnz	kbTranslateDone

	test	bh, KB_STATE_NUM
	jz	kbTranslateDone

	mov	al, [esi+2]

kbTranslateDone:
	clc
	ret

EndProc kbTranslate

;---------------------------------------------------------------------------
;	kbEventHandler
;
;	Entry:
;	      	None
;	Exit:
;	     	None
;
DefProc kbEventHandler

	push	eax
	push	ebx
	push	esi

	cli
	call	kbDisable
	sti

	in	al, 60h

	call	kbCheckFlags		; check for Flags
	jnc	kbEvHndExit		;

	call	kbCheckShiftKeys	; check for Shift Keys
	jnc	kbEvHndExit		;

	call	kbCheckLEDKeys		; check for "LED" Keys
	jnc	kbEvHndExit		;

	cmp	al, KB_SCAN_LAST	; valid keypress?
	ja	kbEvHndDone		;

	call	kbCheckReboot		; check for <CTRL>+<ALT>+<DEL>
	jnc	kbEvHndReboot		;

;	call	kbCheckDeadKeys		; check for "Dead" keys
;	jnc	kbEvHndDone		;

	call	kbCheckAltGr		; check for AltGr keys
	jnc	kbEvHndSave		;

	call	kbCheckExtended		; check for Extended keys
	jnc	kbEvHndSave		;

	call	kbTranslate		; perform IBM INT16 translation
	jc	kbEvHndDone		;

kbEvHndSave:

	; ToDo: either save keycode in a queue 
	;	or call a notification handler

kbEvHndDone:
	and	kbFlags, NOT KB_FLAGS_E0

kbEvHndExit:	
	cli
	call	kbEnable
	sti

	pop	esi
	pop	ebx
	pop	eax

	jmp	kbOldHandler

kbEvHndReboot:
	APIcall	KERNEL_Reboot
	jmp	kbEvHndDone

EndProc kbEventHandler

codeSegment ENDS

;---------------------------------------------------------------------------
; Data
;---------------------------------------------------------------------------

dataSegment SEGMENT

;---------------------------------------------------------------------------
; MINIKBD.SYS API Service Table
;---------------------------------------------------------------------------

apiServiceTbl	LABEL	DWORD
	API_TABLE	<apiServiceDef, (apiServiceEnd-apiServiceDef)/8>
apiServiceDef	LABEL	DWORD
	API_ENTRY	<KEYBOARD_GetVersion,	kbGetVersion>
apiServiceEnd	LABEL	DWORD

;---------------------------------------------------------------------------
; default Keyboard Layout definitions for US keyboard
;---------------------------------------------------------------------------

kbLayout	DD	kbLayoutUS

kbLayoutUS	KB_LAYOUT <"US", NULL, usMain>

; Allow FFh hex without leading zero for neatness.

FFh		EQU	0FFh

		;	+--------------------- 0=Normal
		;	                      1=Normal, Caps-Lock Sensitive
		;	                      2=Zeroed, ASCII=00h
		;	                      3=Numpad
		;	   +----------------- Normal
		;	       +------------ Shift
		;	           +------- Ctrl
		;	               +-- Alt
		;	               

usMain		LABEL	BYTE
	
		DB	0, FFh, FFh, FFh, FFh	; 00 - Error
		DB	0, 1Bh, 1Bh, 1Bh, FFh	; 01 - Escape
		DB	0, "1", "!", FFh, 78h	; 02 - 1
		DB	0, "2", "@", FFh, 79h	; 03 - 2
		DB	0, "3", "#", FFh, 7Ah	; 04 - 3
		DB	0, "4", "$", FFh, 7Bh	; 05 - 4
		DB	0, "5", "%", FFh, 7Ch	; 06 - 5
		DB	0, "6", "^", FFh, 7Dh	; 07 - 6
		DB	0, "7", "&", FFh, 7Eh	; 08 - 7
		DB	0, "8", "*", FFh, 7Fh	; 09 - 8
		DB	0, "9", "(", FFh, 80h	; 0A - 9
		DB	0, "0", ")", FFh, 81h	; 0B - 0
		DB	0, "-", "_", FFh, 82h	; 0C - Minus
		DB	0, "=", "+", FFh, 83h	; 0D - Equal
		DB	0, 08h, 08h, FFh, FFh	; 0E - Backspace
		DB	0, 09h, 00h, FFh, FFh	; 0F - Tab
		DB	1, "q", "Q", 11h, 10h	; 10 - Q
		DB	1, "w", "W", 17h, 11h	; 11 - W
		DB	1, "e", "E", 05h, 12h	; 12 - E
		DB	1, "r", "R", 12h, 13h	; 13 - R
		DB	1, "t", "T", 14h, 14h	; 14 - T
		DB	1, "y", "Y", 19h, 15h	; 15 - Y
		DB	1, "u", "U", 15h, 16h	; 16 - U
		DB	1, "i", "I", 09h, 17h	; 17 - I
		DB	1, "o", "O", 0Fh, 18h	; 18 - O
		DB	1, "p", "P", 10h, 19h	; 19 - P
		DB	0, "[", "{", 1Bh, FFh	; 1A - [
		DB	0, "]", "}", 1Dh, FFh	; 1B - ]
		DB	0, 0Dh, 0Dh, 0Ah, FFh	; 1C - Enter
		DB	0, FFh, FFh, FFh, FFh	; 1D - Ctrl
		DB	1, "a", "A", 01h, 1Eh	; 1E - A
		DB	1, "s", "S", 13h, 1Fh	; 1F - S
		DB	1, "d", "D", 04h, 20h	; 20 - D
		DB	1, "f", "F", 06h, 21h	; 21 - F
		DB	1, "g", "G", 07h, 22h	; 22 - G
		DB	1, "h", "H", 08h, 23h	; 23 - H
		DB	1, "j", "J", 0Ah, 24h	; 24 - J
		DB	1, "k", "K", 0Bh, 25h	; 25 - K
		DB	1, "l", "L", 0Ch, 26h	; 26 - L
		DB	0, ";", ":", FFh, FFh	; 27 - Semicolon
		DB	0, "'","""", FFh, FFh	; 28 - Apostrophe
		DB	0, "`", "~", 1Ch, FFh	; 29 -
		DB	0, FFh, FFh, FFh, FFh	; 2A - Shift
		DB	0, "\", "|", FFh, FFh	; 2B - Slash
		DB	1, "z", "Z", 1Ah, 2Ch	; 2C - Z
		DB	1, "x", "X", 18h, 2Dh	; 2D - X
		DB	1, "c", "C", 03h, 2Eh	; 2E - C
		DB	1, "v", "V", 16h, 2Fh	; 2F - V
		DB	1, "b", "B", 02h, 30h	; 30 - B
		DB	1, "n", "N", 0Eh, 31h	; 31 - N
		DB	1, "m", "M", 0Dh, 32h	; 32 - M
		DB	0, ",", "<", FFh, FFh	; 33 - Comma
		DB	0, ".", ">", FFh, FFh	; 34 - Period
		DB	0, "/", "?", FFh, FFh	; 35 - Slash
		DB	0, FFh, FFh, FFh, FFh	; 36 - Shift
		DB	0, "*", "*", FFh, FFh	; 37 - Asterisks
		DB	0, FFh, FFh, FFh, FFh	; 38 - Alt
		DB	0, " ", " ", " ", FFh	; 39 - Space
		DB	0, FFh, FFh, FFh, FFh	; 3A - Caps
		DB	2, 3Bh, 54h, 5Eh, 68h	; 3B - F1
		DB	2, 3Ch, 55h, 5Fh, 69h	; 3C - F2
		DB	2, 3Dh, 56h, 60h, 6Ah	; 3D - F3
		DB	2, 3Eh, 57h, 61h, 6Bh	; 3E - F4
		DB	2, 3Fh, 58h, 62h, 6Ch	; 3F - F5
		DB	2, 40h, 59h, 63h, 6Dh	; 40 - F6
		DB	2, 41h, 5Ah, 64h, 6Eh	; 41 - F7
		DB	2, 42h, 5Bh, 65h, 6Fh	; 42 - F8
		DB	2, 43h, 5Ch, 66h, 70h	; 43 - F9
		DB	2, 44h, 5Dh, 67h, 71h	; 44 - F10
		DB	0, FFh, FFh, FFh, FFh	; 45 - Num
		DB	0, FFh, FFh, FFh, FFh	; 46 - Scroll
		DB	3, 00h, "7", FFh, FFh	; 47 - 7 Home
		DB	3, 00h, "8", FFh, FFh	; 48 - 8 Up
		DB	3, 00h, "9", FFh, FFh	; 49 - 9 PgUp
		DB	0, "-", "-", FFh, FFh	; 4A - Minus
		DB	3, 00h, "4", FFh, FFh	; 4B - 4 Left
		DB	3, 00h, "5", FFh, FFh	; 4C - 5
		DB	3, 00h, "6", FFh, FFh	; 4D - 6 Right
		DB	0, "+", "+", FFh, FFh	; 4E - Plus
		DB	3, 00h, "1", FFh, FFh	; 4F - 1 End
		DB	3, 00h, "2", FFh, FFh	; 50 - 2 Down
		DB	3, 00h, "3", FFh, FFh	; 51 - 3 PgDn
		DB	3, 00h, "0", FFh, FFh	; 52 - Insert
		DB	3, 00h, ".", FFh, FFh	; 53 - Delete
		DB	0, FFh, FFh, FFh, FFh	; 54 - SysReq
		DB	0, FFh, FFh, FFh, FFh	; 55 -
		DB	0, 5Ch, 7Ch, FFh, FFh	; 56 - Backslash
		DB	2, 85h, 87h, 89h, 8Bh	; 57 - F11
		DB	2, 86h, 88h, 8Ah, 8Ch	; 58 - F12
		DB	0, FFh, FFh, FFh, FFh	; 59 -
		DB	0, FFh, FFh, FFh, FFh	; 5A -
		DB	0, FFh, FFh, FFh, FFh	; 5B - LWin
		DB	0, FFh, FFh, FFh, FFh	; 5C - RWin
		DB	2, 98h, 99h, 9Ah, 9Bh	; 5D - Menu
		DB	0, FFh, FFh, FFh, FFh	; 5E -
		DB	0, FFh, FFh, FFh, FFh	; 5F -

dataSegment ENDS

;---------------------------------------------------------------------------
; BSS Data
;---------------------------------------------------------------------------

bssSegment SEGMENT

kbOldHandler	DD	?

kbFlags		DB	?
kbShift		DB	?
kbState		DB	?

bssSegment ENDS

	END	mainEntryPoint
