;=-------------------------------------------------------------------------=
;
;			SYSDRV - Initialize System Drive
;
;     Copyright (c) 1999-2000 by Thomas Kjoernes, all rights reserved.
;
;		Usage:	SYSDRV <drive:>
;
;=-------------------------------------------------------------------------=

	INCLUDE std.inc
	INCLUDE	dos.inc
	INCLUDE	fat.inc

	.386p

codeSegment SEGMENT USE16

	ASSUME	cs:codeSegment, ds:NOTHING, es:NOTHING

dosEntryPoint:

	; setup our own stack

	push	cs
	pop	ss

	mov	sp, OFFSET ourStackTop

	; grab all command line parameters and put them
	; into our buffer neatly converted to uppercase.

	push	cs
	pop	es

	mov	di, OFFSET ourOwnPath
	call	dosGetStartupPath

	mov	di, OFFSET ourCmdLine
	call	stdParseCmdLine

	push	cs
	pop	ds

	ASSUME	ds:codeSegment

	mov	si, OFFSET msgSysInfo
	call	dosPrintMsgZ

	mov	si, OFFSET msgSysHelp
	jcxz	errTerminate

	; verify the drive letter

	mov	ax, WORD PTR ourCmdLine[0]
	cmp	ax, ":A"
	jb	errDriveLetter
	cmp	ax, ":Z"
	ja	errDriveLetter
	sub	al, "A"
	mov	ourDriveNum, al

	call	sysTransferSystem
	jc	errTerminate

	mov	si, OFFSET msgSysDone
	call	dosPrintMsgZ

dosTerminate:
	mov	ax, 4C00h
	int	21h

	; invalid drive letter, show error about that.

errDriveLetter:
	mov	si, OFFSET msgErrDrive
	call	dosPrintMsgZ
	mov	si, OFFSET ourCmdLine
	call	dosPrintMsgZ
	mov	si, OFFSET msgCRLF

errTerminate:
	call	dosPrintMsgZ

	mov	ax, 4CFFh
	int	21h

;---------------------------------------------------------------------------
;	sysTransferSystem - 
;
;	Entry:
;		al = drive number
;
sysTransferSystem PROC NEAR

	; complete destination path

	add	newSysPath[0], al

	; combine new path with loader name

	mov	di, OFFSET newSysPath + 3
	mov	si, OFFSET ourSysFile
	call	stdCopyString

	; combine our own path with loader name

	mov	si, OFFSET ourOwnPath
	mov	di, OFFSET ourSysPath
	call	stdGetPathString

	mov	si, OFFSET ourSysFile
	call	stdCopyString

	call	sysCopyLoader
	jc	sysTransferExit

	call	sysUpdateBoot

sysTransferExit:
	ret

sysTransferSystem ENDP

;---------------------------------------------------------------------------
;	sysUpdateBoot - 
;
;	Entry:
;		None
;
sysUpdateBoot PROC NEAR

	mov	si, OFFSET msgErrBootRec

	push	si

	mov	al, ourDriveNum
	xor	ebx, ebx
	mov	cx, 1
	mov	dx, OFFSET oldBootRec
	call	dosDiskRead
	jc	sysUpdateExit

	; check for FAT12 boot sector

	mov	si, OFFSET oldBootRec.bsFileSystem
	mov	di, OFFSET strFAT12
	mov	cx, 4
	rep	cmpsw
	je	sysUpdateMain

	; check for FAT16 boot sector

	mov	si, OFFSET oldBootRec.bsFileSystem
	mov	di, OFFSET strFAT16
	mov	cx, 4
	rep	cmpsw
	je	sysUpdateMain

	jmp	sysUpdateFail

sysUpdateMain:	
	push	di
	mov	si, OFFSET ourOwnPath
	mov	di, OFFSET newRecPath
	call	stdGetPathString
	pop	si
	call	stdCopyString

	mov	dx, OFFSET newRecPath
	mov	al, OPEN_READ_ONLY
	call	dosFileOpen
	jc	sysUpdateFail

	xchg	ax, bx
	mov	cx, SIZE newBootRec
	mov	dx, OFFSET newBootRec
	call	dosFileRead
	call	dosFileClose

	mov	si, OFFSET oldBootRec.bsFAT12
	mov	di, OFFSET newBootRec.bsFAT12

	cld
	mov	cx, SIZE bsFAT12
	rep	movsb

	; write out new boot sector

	mov	al, ourDriveNum
	xor	ebx, ebx
	mov	cx, 1
	mov	dx, OFFSET newBootRec
	call	dosDiskWrite

	clc
	jmp	sysUpdateExit

sysUpdateFail:
	stc

sysUpdateExit:
	pop	si
	ret

sysUpdateBoot ENDP

;---------------------------------------------------------------------------
;	sysCopyLoader - 
;
;	Entry:
;		None
;
sysCopyLoader PROC NEAR

	push	ax
	push	bx
	push	cx
	push	dx
	push	bp

	; attempt to open loader

	mov	si, OFFSET msgErrOpen
	mov	dx, OFFSET ourSysPath
	mov	al, OPEN_READ_ONLY
	call	dosFileOpen
	jc	sysCopyLdrDone

	xchg	ax, bx

	; attempt to create copy of loader on
	; the specifed destination drive...

	mov	si, OFFSET msgErrCreate
	mov	dx, OFFSET newSysPath
	call	dosFileCreate
	jc	sysCopyLdrDone

	xchg	ax, bp

sysCopyLdrNext:
	mov	ah, 3Fh
	mov	cx, SIZE ourCopyBuf
	mov	dx, OFFSET ourCopyBuf
	int	21h

	xchg	ax, cx
	jcxz	sysCopyLdrDone

	push	bx
	mov	ah, 40h
	mov	bx, bp
	int	21h
	pop	bx

	jmp	sysCopyLdrNext

sysCopyLdrDone:
	pop	bp
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	ret

sysCopyLoader ENDP

;---------------------------------------------------------------------------
;	stdParseCmdLine - 
;
;	Entry:
;		ds = PSP segment
;	     es:di = buffer for "formatted" parameters
;
;	Exit:
;
stdParseCmdLine PROC NEAR

	push	ax
	push	si
	push	di

	xor	cx, cx		; clear argument counter
	mov	si, 128 + 1	; point to command line string

stdParseSkip:
	lodsb			; get command line character

	cmp	al, 13		; terminating CR?
	je	stdParseExit	;
	cmp	al, " "		; space?
	je	stdParseSkip	;

	inc	cx

stdParseMore:
	cmp	al, "a"		; is it too low for conversion?
	jb	stdParseSave	;
	cmp	al, "z"		; is it too high for conversion?
	ja	stdParseSave	;
	sub	al, " "		; in range a..z, make it A..Z

stdParseSave:
	stosb			; copy this character
	lodsb			; also get the next one

	cmp	al, 13		; terminating CR?
	je	stdParseExit	;
	cmp	al, " "		; space?
	jne	stdParseMore	;

	xor	al, al		; we have another string... store
	stosb			; null terminator for previous string
	jmp	stdParseSkip	; and continue with next string

stdParseExit:
	xor	ax, ax		; store double null terminator for
	stosw			; entire list of strings

	pop	di
	pop	si
	pop	ax
	ret

stdParseCmdLine ENDP

;---------------------------------------------------------------------------
;	stdCopyString - Copies a null terminated string into a buffer.
;
;	Entry:
;
;	     ds:si = pointer to source string
;	     es:di = pointer to target buffer
;
;	Exit:
;
;	     es:di = points to null terminator
;
stdCopyString PROC NEAR

	push	ax
	push	si

stdCopyStrNext:
	lodsb			; 
	stosb			; copy string
	or	al, al		;
	jnz	stdCopyStrNext	; until null terminator is found

	dec	di		; point back to NULL

	pop	si
	pop	ax
	ret

stdCopyString ENDP

;---------------------------------------------------------------------------
;	stdGetPathString - Copies the path to a buffer, excluding filename.
;
;	Entry:
;
;	     ds:si = pointer to source string
;	     es:di = pointer to target buffer
;
;	Exit:
;
;	     es:di = points to null terminator
;
stdGetPathString PROC NEAR

	push	ax
	push	bx
	push	cx
	push	si

	cmp	BYTE PTR [si][1], ":"
	jne	stdGetPathStrMain

	movsw			; store drive specifier
	
stdGetPathStrMain:
	mov	bx, si		; save start pointer
	xor	cx, cx		; zero "\" pointer

stdGetPathStrNext:
	lodsb
	or	al, al
	jz	stdGetPathStrCopy
	cmp	al, "\"
	jne	stdGetPathStrNext

	; we have a directory separator, update "\" pointer

	mov	cx, si
	jmp	stdGetPathStrNext

stdGetPathStrCopy:
	jcxz	stdGetPathStrExit

	mov	si, bx		; pointer start of string
	sub	cx, bx		; calculate path length
	rep	movsb		; copy it!

stdGetPathStrExit:
	xor	al, al		;
	stosb			; store null terminator
	dec	di		; point back to NULL

	pop	si
	pop	cx
	pop	bx
	pop	ax
	ret

stdGetPathString ENDP

;---------------------------------------------------------------------------
;	dosLockVolume - 
;
;	Entry:
;		al = DOS drive number (0=A, 1=B, etc...)
;	Exit:
;		CF=1 if unable to LOCK volume.
;
dosLockVolume PROC NEAR

	pusha
	inc	al		; make drive number one-based
	mov	bl, al		; move into BL
	mov	bh, 00h		; lock type -- 00
	mov	ax, 440Dh	; generic block device IOCTL
	mov	cx, 084Ah	; code 08h (disk), subcode 4Ah (lock)
	mov	dx, 0001h	; permission -- allow writes
	int	21h		; invoke DOS
	popa
	ret

dosLockVolume ENDP

;---------------------------------------------------------------------------
;	dosUnlockVolume - 
;
;	Entry:
;		al = DOS drive number (0=A, 1=B, etc...)
;	Exit:
;		CF=1 if unable to UNLOCK volume.
;
dosUnlockVolume PROC NEAR

	pusha
	inc	al		; make drive number one-based
	mov	bl, al		; move into BL
	mov	ax, 440Dh	; generic block device IOCTL
	mov	cx, 086Ah	; code 08h (disk), subcode 4Ah (lock)
	int	21h		; invoke DOS
	popa			
	ret

dosUnlockVolume ENDP

;---------------------------------------------------------------------------
;	dosDiskRead - 
;
;	Entry:
;		al = DOS drive number (0=A, 1=B, etc...)
;		cx = sector count
;	       ebx = sector #
;	     ds:dx = buffer address
;
dosDiskRead PROC NEAR

	pusha
	push	ds

	push	ds		; buffer SEG
	push	dx		; buffer OFS
	push	cx		; sector count
	push	ebx		; sector LBA #
	mov	bx, sp		; sp -> packet

	push	ss		; our packet is on the stack,
	pop	ds		; but DOS wants ds:dx as pointer

	mov	cx, -1		; specify extended read
	int	25h		; absolute disk read!
	pop	ax		; pop flags, long story...

	add	sp, 10		; size of disk I/O packet
	pop	ds
	popa
	ret

dosDiskRead ENDP

;---------------------------------------------------------------------------
;	dosDiskWrite - 
;
;	Entry:
;		al = DOS drive number (0=A, 1=B, etc...)
;		cx = sector count
;	       ebx = sector #
;	     ds:dx = buffer address
;
dosDiskWrite PROC NEAR

	call	dosLockVolume

	pusha
	push	ds

	push	ds		; buffer SEG
	push	dx		; buffer OFS
	push	cx		; sector count
	push	ebx		; sector LBA #
	mov	bx, sp		; sp -> packet

	push	ss		; our packet is on the stack,
	pop	ds		; but DOS wants ds:dx as pointer

	mov	cx, -1		; specify extended write
	int	26h		; absolute disk Write!
	pop	ax		; pop flags, long story...

	add	sp, 10		; size of disk I/O packet
	pop	ds
	popa

	call	dosUnlockVolume
	
	ret

dosDiskWrite ENDP

;---------------------------------------------------------------------------
;	dosFileCreate - 
;
;	Entry:
;	
;	Exit:
;
dosFileCreate PROC NEAR

	mov	ah, 3Ch
	xor	cx, cx
	int	21h
	ret

dosFileCreate ENDP

;---------------------------------------------------------------------------
;	dosFileOpen - 
;
;	Entry:
;	
;	Exit:
;
dosFileOpen PROC NEAR

	mov	ah, 3Dh
	int	21h
	ret

dosFileOpen ENDP

;---------------------------------------------------------------------------
;	dosFileClose - 
;
;	Entry:
;	
;	Exit:
;
dosFileClose PROC NEAR

	mov	ah, 3Eh
	int	21h
	ret

dosFileClose ENDP

;---------------------------------------------------------------------------
;	dosFileRead - 
;
;	Entry:
;	
;	Exit:
;
dosFileRead PROC NEAR

	mov	ah, 3Fh
	int	21h
	ret

dosFileRead ENDP

;---------------------------------------------------------------------------
;	dosPrintChar -
;
;	Entry:
;		al = character
;
dosPrintChar PROC NEAR

	push	ax
	push	dx

	mov	ah, 02h
	mov	dl, al
	int	21h

	pop	dx
	pop	ax
	ret

dosPrintChar ENDP

;---------------------------------------------------------------------------
;	dosPrintCRLF -
;
;	Entry:
;		None
;
dosPrintCRLF PROC NEAR

	push	ax
	mov	al, 13		; output ASCII 13 - CR
	call	dosPrintChar	;
	mov	al, 10		; output ASCII 10 - LF
	call	dosPrintChar	;
	pop	ax
	ret

dosPrintCRLF ENDP

;---------------------------------------------------------------------------
;	dosPrintMsgZ -
;
;	Entry:
;
;	     ds:si = pointer to message
;
dosPrintMsgZ PROC NEAR

	push	ax
	push	si

dosPMZNext:
	lodsb			; get character

	or	al, al		; null?
	jz	dosPMZDone	;

	call	dosPrintChar	; print charcater using BIOS service
	jmp	dosPMZNext	;

dosPMZDone:
	pop	si
	pop	ax
	ret

dosPrintMsgZ ENDP

;---------------------------------------------------------------------------
;	dosGetStartupPath - 
;
;	Entry:
;		ds = PSP
;	     es:di = pointer to buffer for startup path
;
dosGetStartupPath PROC NEAR

	push	ax
	push	si
	push	di
	push	ds

	mov	ds, ds:[002Ch]
	xor	si, si

dosGetStartupPathNext:
	cmp	WORD PTR [si], NULL
	je	dosGetStartupPathMain
	inc	si
	jns	dosGetStartupPathNext

dosGetStartupPathMain:
	add	si, 4

dosGetStartupPathCopy:
	lodsb
	stosb
	or	al, al
	jnz	dosGetStartupPathCopy
	
	pop	ds
	pop	di
	pop	si
	pop	ax
	ret

dosGetStartupPath ENDP

;---------------------------------------------------------------------------
; MESSAGES
;---------------------------------------------------------------------------

msgCRLF		DB	13, 10, NULL

msgSysInfo	DB	"System Drive Initialization Utility v1.0a [RM] ", BUILD, 13, 10
		DB	"Copyright (c) 1999-2000 by Thomas Kjoernes, "
		DB	"all rights reserved.", 13, 10, NULL

msgSysHelp	DB	13, 10
		DB	"Usage: SYSDRV [drive:]", 13, 10, NULL

msgSysDone	DB	13, 10
		DB	"Successful!", 13, 10, NULL

msgErrBootRec	DB	13, 10
		DB	"Error updating the Boot Record!", 13, 10, NULL

msgErrDrive	DB	13, 10
		DB	"Invalid drive specifier -- ", NULL

msgErrCreate	DB	13, 10
		DB	"Unable to create copy of LOADER.EXE "
		DB	"on destination drive!", 13, 10, NULL

msgErrOpen	DB	13, 10
		DB	"Unable to locate LOADER.EXE file!", 13, 10, NULL

;---------------------------------------------------------------------------
; VARIABLES
;---------------------------------------------------------------------------

ourDriveNum	DB	-1

ourSysFile	DB	"LOADER.EXE", NULL

strFAT12	DB	"FAT12   ", "FAT12.BIN", NULL
strFAT16	DB	"FAT16   ", "FAT16.BIN", NULL

newSysPath	DB	"A:\"
		DB	125 DUP (?)

newRecPath	DB	128 DUP (?)
ourSysPath	DB	128 DUP (?)
ourOwnPath	DB	128 DUP (?)
ourCmdLine	DB	128 DUP (?)

oldBootRec	DB	512 DUP (?)
newBootRec	DB	512 DUP (?)

ourCopyBuf	DB	16384 DUP (?)

;---------------------------------------------------------------------------
; STACK AREA
;---------------------------------------------------------------------------

ourStackDef 	LABEL	WORD
		DB	2048 DUP (?)
ourStackTop 	LABEL	WORD

codeSegment ENDS

	END	dosEntryPoint
