;---------------------------------------------------------------------------
;
; 		    32-bit PM Generic VGA Display Driver
;
;	Copyright (c) 1999 by Thomas Kjoernes, all rights reserved.
;
;---------------------------------------------------------------------------

	INCLUDE std.inc
	INCLUDE stdgroup.inc
	INCLUDE kernel.api
	INCLUDE	vga.api

	.386p

	; Publics

codeSegment SEGMENT

;---------------------------------------------------------------------------
;	vgaGetVersion - 
;
;	Entry:
;		None
;	Exit:
;		eax = version ID
;
vgaGetVersion PROC NEAR

	mov	eax, 00000100h
	clc
	ret

vgaGetVersion ENDP

;---------------------------------------------------------------------------
;	vgaGetTextBuf
;
;	Entry:
;		None
;	Exit:
;	       edi = Text Buffer Address
;		bl = Text Buffer Width
;		bh = Text Buffer Height
;
vgaGetTextBuf PROC NEAR

	mov	edi, ourTextBuf
	mov	bl, ourBufW
	mov	bh, ourBufH
	ret

vgaGetTextBuf ENDP

;---------------------------------------------------------------------------
;	vgaSetTextBuf
;
;	Entry:
;	       eax = Text Buffer Address
;		bl = Text Buffer Width
;		bh = Text Buffer Height
;	Exit:
;		None
;
vgaSetTextBuf PROC NEAR

	mov	ourTextBuf, eax
	mov	ourBufW, bl
	mov	ourBufH, bh
	ret

vgaSetTextBuf ENDP

;---------------------------------------------------------------------------
;	vgaGetFontBuf
;
;	Entry:
;		None
;	Exit:
;	       edi = Font Buffer Address
;
vgaGetFontBuf PROC NEAR

	mov	edi, ourFontBuf
	ret

vgaGetFontBuf ENDP

;---------------------------------------------------------------------------
;	vgaGetVideoMode - 
;
;	Entry:
;		None
;	Exit:
;		eax = Current Video Mode
;
vgaGetVideoMode PROC NEAR

	ret

vgaGetVideoMode ENDP

;---------------------------------------------------------------------------
;	vgaSetVideoMode - 
;
;	Entry:
;		eax = New Video Mode
;	Exit:
;		None
;
vgaSetVideoMode PROC NEAR

	ret

vgaSetVideoMode ENDP

;---------------------------------------------------------------------------
;	vgaFontAccessEnable
;
;	Entry:
;		None
;	Exit:
;		None
;
vgaFontAccessEnable PROC NEAR

	push	eax
	push	edx

	mov     dx, 03C4h
	mov     ax, 0402h
	out     dx, ax
	mov     ax, 0704h
	out     dx, ax
	mov     dx, 03CEh
	mov     ax, 0005h
	out     dx, ax
	mov     ax, 0406h
	out     dx, ax
	mov     ax, 0204h
	out     dx, ax

	pop	edx
	pop	eax
	ret

vgaFontAccessEnable ENDP

;---------------------------------------------------------------------------
;	vgaFontAccessDisable
;
;	Entry:
;		None
;	Exit:
;		None
;
vgaFontAccessDisable PROC NEAR

	push	eax
	push	edx

	mov     dx, 03C4h
	mov     ax, 0302h
	out     dx, ax
	mov     ax, 0304h
	out     dx, ax
	mov     dx, 03CEh
	mov     ax, 1005h
	out     dx, ax
	mov     ax, 0E06h
	out     dx, ax
	mov     ax, 0004h
	out     dx, ax

	pop	edx
	pop	eax
	ret

vgaFontAccessDisable ENDP

;---------------------------------------------------------------------------
;	vgaGetTextFont
;
;	Entry:
;		al = TextFont number
;		cl = TextFont height
;	       edi = TextFont buffer pointer
;	Exit:
;		al = TextFont number (increased by 1)
;	       edi = TextFont buffer pointer (updated beyond TextFont)
;
vgaGetTextFont PROC NEAR

	push	ecx
	push	esi

	call	vgaFontAccessEnable

	movzx	ecx, cl
	movzx	esi, al
	shl	esi, 5
	add	esi, ourFontBuf
	cld
	rep	movsb

	call	vgaFontAccessDisable

	inc	al

	pop	esi
	pop	ecx
	ret

vgaGetTextFont ENDP

;---------------------------------------------------------------------------
;	vgaSetTextFont
;
;	Entry:
;		al = TextFont number
;		cl = TextFont height
;	       esi = TextFont buffer pointer
;	Exit:
;		al = TextFont number (increased by 1)
;	       esi = TextFont buffer pointer (updated beyond TextFont)
;
vgaSetTextFont PROC NEAR


	push	ecx
	push	edi

	call	vgaFontAccessEnable

	movzx	ecx, cl
	movzx	edi, al
	shl	edi, 5
	add	edi, ourFontBuf
	cld
	rep	movsb

	call	vgaFontAccessDisable

	inc	al

	pop	edi
	pop	ecx
	ret

vgaSetTextFont ENDP

;---------------------------------------------------------------------------
;	vgaGetPalette
;
;	Entry:
;		al = Palette Entry Start
;	       ecx = Palette Entry Count
;	       edi = Pointer to RGB data buffer
;	Exit:
;		None
;
vgaGetPalette PROC NEAR

	push	ecx
	push	edx
	push	esi

	mov	edx, 03C7h
	out	dx, al
	inc	edx
	inc	edx
	lea	ecx, [ecx+ecx*2]
	rep	insb

	pop	esi
	pop	edx
	pop	ecx
	ret

vgaGetPalette ENDP

;---------------------------------------------------------------------------
;	vgaSetPalette
;
;	Entry:
;		al = Palette Entry Start
;	       ecx = Palette Entry Count
;	       esi = Pointer to RGB data	
;	Exit:
;		None
;
vgaSetPalette PROC NEAR

	push	ecx
	push	edx
	push	esi

	mov	edx, 03C8h
	out	dx, al
	inc	edx
	lea	ecx, [ecx+ecx*2]
	rep	outsb

	pop	esi
	pop	edx
	pop	ecx
	ret

vgaSetPalette ENDP

;---------------------------------------------------------------------------
;	vgaGetCursorPos
;
;	Entry:
;		None
;	Exit:
;		dl = Cursor X-Position
;		dh = Cursor Y-Position
;
vgaGetCursorPos PROC NEAR

	mov	dl, ourCursorX
	mov	dh, ourCursorY
	ret

vgaGetCursorPos ENDP

;---------------------------------------------------------------------------
;	vgaSetCursorPos
;
;	Entry:
;		dl = New Cursor X-Position
;		dh = New Cursor Y-Position
;	Exit:
;		None
;
vgaSetCursorPos PROC NEAR

	push	eax
	push	ebx
	push	edx

	mov	ourCursorX, dl
	mov	ourCursorY, dh

	; Hack: sync BIOS cursor position

	mov	WORD PTR ds:[0450h], dx

	movzx	ebx, dl
	mov	al, ourBufW
	mul	dh

	add	ebx, eax

	mov	dx, 03D4h

	mov	al, 0Eh
	mov	ah, bh
	out	dx, ax

	mov	al, 0Fh
	mov	ah, bl
	out	dx, ax

	pop	edx
	pop	ebx
	pop	eax
	ret

vgaSetCursorPos ENDP

;---------------------------------------------------------------------------
;	vgaGetCursorShape
;
;	Entry:
;		None
;	Exit:
;		cl = Cursor Starting Scanline (Bit5=1 for Disable)
;		ch = Cursor Ending Scanline
;
vgaGetCursorShape PROC NEAR

	mov	cl, ourCursorA
	mov	ch, ourCursorB
	ret

vgaGetCursorShape ENDP

;---------------------------------------------------------------------------
;	vgaSetCursorShape
;
;	Entry:
;		cl = Cursor Starting Scanline (Bit5=1 for Disable)
;		ch = Cursor Ending Scanline
;	Exit:
;		None
;
vgaSetCursorShape PROC NEAR

	push	eax
	push	edx

	mov	ourCursorA, cl
	mov	ourCursorB, ch

	mov	dx, 03D4h
	mov	al, 0Ah
	mov	ah, cl
	out	dx, ax
	mov	al, 0Bh
	mov	ah, ch
	out	dx, ax

	pop	edx
	pop	eax
	ret

vgaSetCursorShape ENDP

;---------------------------------------------------------------------------
;	vgaWriteChar
;
;	Entry:
;		al = Character
;		ah = Attribute
;		cl = X
;		dl = Y
;	Exit:
;		None
;
vgaWriteChar PROC NEAR

	push	ebx
	push	ecx
	push	edx

	cmp	cl, ourBufW		; outside buffer edges?
	jae	vgaWriteCharExit	;

	cmp	dl, ourBufH		; outside buffer bottom/top?
	jae	vgaWriteCharExit	;

	push	eax
	
	xor	eax, eax
	xor	ebx, ebx

	mov	bl, cl			; bl = Abs.X
	mov	al, dl			; al = Abs.Y

	mul	ourBufW			; al = Abs.Y * Buffer.Width

	add	ebx, eax		; ebx = Abs.X + Abs.Y * ourBufW
	shl	ebx, 1			; *2

	add	ebx, ourTextBuf		; add display memory base

	pop	eax

	mov	[ebx], ax

vgaWriteCharExit:
	pop	edx
	pop	ecx
	pop	ebx
	ret

vgaWriteChar ENDP

;---------------------------------------------------------------------------
;	vgaWriteAttr
;
;	Entry:
;		al = Attribute
;		cl = X
;		dl = Y
;	Exit:
;		None
;
vgaWriteAttr PROC NEAR

	push	ebx
	push	ecx
	push	edx

	cmp	cl, ourBufW		; outside buffer edges?
	jae	vgaWriteAttrExit	;

	cmp	dl, ourBufH		; outside buffer bottom/top?
	jae	vgaWriteAttrExit	;

	push	eax
	
	xor	eax, eax
	xor	ebx, ebx

	mov	bl, cl			; bl = Abs.X
	mov	al, dl			; al = Abs.Y

	mul	ourBufW			; al = Abs.Y * Buffer.Width

	add	ebx, eax		; ebx = Abs.X + Abs.Y * ourBufW
	shl	ebx, 1			; *2

	add	ebx, ourTextBuf		; add display memory base

	pop	eax

	mov	[ebx], al

vgaWriteAttrExit:
	pop	edx
	pop	ecx
	pop	ebx
	ret

vgaWriteAttr ENDP

;---------------------------------------------------------------------------
;	vgaBlockChar
;
;	Entry:
;		al = Character
;		ah = Attribute
;		cl = X
;		ch = Width
;		dl = Y
;		dh = Height
;	Exit:
;		None
;
vgaBlockChar PROC NEAR

	push	ebx
	push	edx

	mov	ebx, ecx

vgaBlockCharNext:
	call	vgaWriteChar
	inc	cl
	dec	ch
	jnz	vgaBlockCharNext

	mov	ecx, ebx

	inc	dl
	dec	dh
	jnz	vgaBlockCharNext

	pop	edx
	pop	ebx
	ret

vgaBlockChar ENDP

;---------------------------------------------------------------------------
;	vgaBlockAttr
;
;	Entry:
;		al = Attribute
;		cl = X
;		ch = Width
;		dl = Y
;		dh = Height
;	Exit:
;		None
;
vgaBlockAttr PROC NEAR

	push	ebx
	push	edx

	mov	ebx, ecx

vgaBlockAttrNext:
	call	vgaWriteAttr
	inc	cl
	dec	ch
	jnz	vgaBlockAttrNext

	mov	ecx, ebx

	inc	dl
	dec	dh
	jnz	vgaBlockAttrNext

	pop	edx
	pop	ebx
	ret

vgaBlockAttr ENDP

;---------------------------------------------------------------------------
;	mainEntryPoint - VGA.SYS entry point
;
;	Entry (Stack):
;
;		DWORD = return address
;
;	Entry (Registers):
;
;		ebx = our memory block base -- "module handle"
;		ecx = our memory block size
;
mainEntryPoint PROC NEAR

	mov	esi, OFFSET apiServiceTbl
	APIcall	KERNEL_ApiRegister

	; sync with BIOS cursor position

	movzx	edx, WORD PTR ds:[0450h]
	APIcall	VGA_SetCursorPos

	clc
	ret

mainEntryPoint ENDP

codeSegment ENDS

;---------------------------------------------------------------------------
; Data
;---------------------------------------------------------------------------

dataSegment SEGMENT

;---------------------------------------------------------------------------
; VGA.SYS API Service Table
;---------------------------------------------------------------------------

apiServiceTbl	LABEL	DWORD
	API_TABLE	<apiServiceDef, (apiServiceEnd-apiServiceDef)/8>
apiServiceDef	LABEL	DWORD
	API_ENTRY	<VGA_GetVersion,	vgaGetVersion>
	API_ENTRY	<VGA_GetTextBuf,	vgaGetTextBuf>
	API_ENTRY	<VGA_SetTextBuf,	vgaSetTextBuf>
	API_ENTRY	<VGA_GetFontBuf,	vgaGetFontBuf>
	API_ENTRY	<VGA_GetVideoMode,	vgaGetVideoMode>
	API_ENTRY	<VGA_SetVideoMode,	vgaSetVideoMode>
	API_ENTRY	<VGA_FontAccessEnable,	vgaFontAccessEnable>
	API_ENTRY	<VGA_FontAccessDisable,	vgaFontAccessDisable>
	API_ENTRY	<VGA_GetTextFont,	vgaGetTextFont>
	API_ENTRY	<VGA_SetTextFont,	vgaSetTextFont>
	API_ENTRY	<VGA_GetCursorPos,	vgaGetCursorPos>
	API_ENTRY	<VGA_SetCursorPos,	vgaSetCursorPos>
	API_ENTRY	<VGA_GetCursorShape,	vgaGetCursorShape>
	API_ENTRY	<VGA_SetCursorShape,	vgaSetCursorShape>
	API_ENTRY	<VGA_GetPalette,	vgaGetPalette>
	API_ENTRY	<VGA_SetPalette,	vgaSetPalette>
	API_ENTRY	<VGA_WriteChar,		vgaWriteChar>
	API_ENTRY	<VGA_WriteAttr,		vgaWriteAttr>
	API_ENTRY	<VGA_BlockChar,		vgaBlockChar>
	API_ENTRY	<VGA_BlockAttr,		vgaBlockAttr>
apiServiceEnd	LABEL	DWORD

ourFontBuf	DD	0A0000h
ourTextBuf	DD	0B8000h

ourBufW		DB	80
ourBufH		DB	25

ourCursorX	DB	0
ourCursorY	DB	0
ourCursorA	DB	0Ch
ourCursorB	DB	0Eh

dataSegment ENDS

;---------------------------------------------------------------------------
; BSS Data
;---------------------------------------------------------------------------

bssSegment SEGMENT

bssSegment ENDS

	END	mainEntryPoint
